# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Twitter plugin for Kegbot."""

from django.conf import settings
from pykeg.plugin import plugin
from pykeg.plugin import util

import tweepy

from kegbot.util import kbjson
from pykeg.core.models import SiteSettings

from . import forms
from . import tasks
from . import views

from unicodedata import normalize

KEY_CONSUMER_KEY = 'consumer_key'
KEY_CONSUMER_SECRET = 'consumer_secret'

KEY_SITE_PROFILE = 'site_profile'
KEY_OAUTH_TOKEN = 'oauth_token'
KEY_OAUTH_TOKEN_SECRET = 'oauth_token_secret'
KEY_TWITTER_NAME = 'twitter_name'
KEY_TWITTER_ID = 'twitter_id'

TWITTER_LINK_LEN = 22
TWEET_MAX_LEN = 140
TWEET_MAX_LEN_WITH_URL = TWEET_MAX_LEN - TWITTER_LINK_LEN - 1
TRUNCATE_STR = u'\u2026'

def get_api(oauth_token, oauth_token_secret):
    auth = tweepy.OAuthHandler(_CONSUMER_KEY, _CONSUMER_SECRET)
    auth.set_access_token(oauth_token, oauth_token_secret)
    return tweepy.API(auth)

def send_tweet(api, contents):
    """Sends a tweet using `api`."""
    return api.update_status(contents)

def truncate_tweet(tweet, max_len=TWEET_MAX_LEN, truncate_str=TRUNCATE_STR):
    if len(tweet) <= max_len:
        return tweet
    words = tweet.strip().split()
    while True:
        tweet = u' '.join(words)
        candidate = normalize('NFC', tweet + truncate_str)
        if not tweet or len(candidate) <= max_len:
            return candidate
        words = words[:-1]


class TwitterPlugin(plugin.Plugin):
    NAME = 'Twitter'
    SHORT_NAME = 'twitter'
    DESCRIPTION = 'Tweet when you pour!'
    URL = 'http://kegbot.org'
    VERSION = '0.0.1-pre'

    def get_admin_settings_view(self):
        return views.admin_settings

    def get_user_settings_view(self):
        return views.user_settings

    def get_extra_admin_views(self):
        return [
            ('redirect/$', 'pykeg.contrib.twitter.views.site_twitter_redirect', 'site_twitter_redirect'),
            ('callback/$', 'pykeg.contrib.twitter.views.site_twitter_callback', 'site_twitter_callback'),
        ]

    def get_extra_user_views(self):
        return [
            ('redirect/$', 'pykeg.contrib.twitter.views.user_twitter_redirect', 'user_twitter_redirect'),
            ('callback/$', 'pykeg.contrib.twitter.views.user_twitter_callback', 'user_twitter_callback'),
        ]

    def handle_new_event(self, event):
        self.logger.info('Handling new event: %s' % event.id)
        if util.is_stale(event.time):
          self.logger.info('Event is stale, ignoring: %s' % event.id)
          return

        profile = self.get_site_profile()
        if not profile:
            self.logger.info('No site twitter profile, ignoring.')
            return

        settings = self.get_saved_form_data(forms.SiteSettingsForm(), 'site_settings')

        self._issue_system_tweet(event, settings, profile)
        if event.user:
            self._issue_user_tweet(event, settings)

    ### Twitter-specific methods

    def get_site_profile(self):
        return self._get_profile(KEY_SITE_PROFILE)

    def get_user_profile(self, user):
        return self._get_profile('user_profile:%s' % user.id)

    def _get_profile(self, datastore_key):
        s = self.datastore.get(datastore_key)
        if s:
            return kbjson.loads(s)
        return {}

    def save_site_profile(self, oauth_token, oauth_token_secret,
            twitter_name, twitter_id):
        return self._save_profile(KEY_SITE_PROFILE, oauth_token,
            oauth_token_secret, twitter_name, twitter_id)

    def save_user_profile(self, user, oauth_token, oauth_token_secret,
            twitter_name, twitter_id):
        return self._save_profile('user_profile:%s' % user.id, oauth_token,
            oauth_token_secret, twitter_name, twitter_id)

    def _save_profile(self, datastore_key, oauth_token, oauth_token_secret,
            twitter_name, twitter_id):
        profile = {
            KEY_OAUTH_TOKEN: oauth_token,
            KEY_OAUTH_TOKEN_SECRET: oauth_token_secret,
            KEY_TWITTER_NAME: twitter_name,
            KEY_TWITTER_ID: twitter_id,
        }
        self.datastore.set(datastore_key, kbjson.dumps(profile))
        return profile

    def remove_site_profile(self):
        self.datastore.delete(KEY_SITE_PROFILE)

    def remove_user_profile(self, user):
        self.datastore.delete('user_profile:%s' % user.id)

    def get_credentials(self):
        if getattr(settings, 'EMBEDDED', False):
            return (
                getattr(settings, 'TWITTER_CONSUMER_KEY', ''),
                getattr(settings, 'TWITTER_CONSUMER_SECRET_KEY', ''),
            )
        return self.datastore.get(KEY_CONSUMER_KEY), self.datastore.get(KEY_CONSUMER_SECRET)

    def get_api(self):
        return get_api(*self.get_credentials())

    def set_credentials(self, consumer_key, consumer_secret):
        self.datastore.set(KEY_CONSUMER_KEY, consumer_key)
        self.datastore.set(KEY_CONSUMER_SECRET, consumer_secret)

    def get_user_settings_form(self, user):
        form = forms.UserSettingsForm()
        self.load_form_defaults(form, 'user_settings:%s' % user.id)
        return form

    def save_user_settings_form(self, user, form):
        self.save_form(form, 'user_settings:%s' % user.id)

    def _issue_system_tweet(self, event, settings, site_profile):
        kind = event.kind
        tweet = None
        append_url = bool(settings.get('append_url'))

        if kind == 'drink_poured':
            if not settings.get('tweet_drink_events'):
                self.logger.info('Skipping system tweet for drink event %s: disabled by settings.' % event.id)
                return
            elif not event.user and not settings.get('include_guests'):
                self.logger.info('Skipping system tweet for event %s: guest pour.' % event.id)
                return
            template = settings.get('drink_poured_template')
            tweet = self._compose_tweet(event, template, append_url)

        elif kind == 'session_started':
            if not settings.get('tweet_session_events'):
                self.logger.info('Skipping system tweet for session start event %s: disabled by settings.' % event.id)
                return
            template = settings.get('session_started_template')
            tweet = self._compose_tweet(event, template, append_url)

        elif kind == 'session_joined':
            if not settings.get('tweet_session_events'):
                self.logger.info('Skipping system tweet for session join event %s: disabled by settings.' % event.id)
                return
            template = settings.get('session_joined_template')
            append_url = bool(settings.get('append_url'))
            tweet = self._compose_tweet(event, template, append_url)

        if tweet:
            self._schedule_tweet(tweet, site_profile)

    def _issue_user_tweet(self, event, settings):
        kind = event.kind
        user = event.user
        tweet = None
        append_url = bool(settings.get('append_url'))

        if kind != 'drink_poured':
            return

        profile = self.get_user_profile(user)
        if not profile:
            self.logger.info('Skipping user tweet for event %s: no twitter profile.' % event.id)
            return
        user_settings = self.get_saved_form_data(forms.SiteSettingsForm(), 'site_settings')

        if event.drink.shout:
            tweet = '%s #kegbot' % event.drink.shout
        else:
            template = settings.get('user_drink_poured_template')
            tweet = self._compose_tweet(event, template, append_url)

        if tweet:
            self._schedule_tweet(tweet, profile)

    def _compose_tweet(self, event, template, append_url):
        kbvars = self.get_vars(event)
        tweet = self.expand_template(template, kbvars)
        if append_url:
            url = kbvars['url']
            tweet = truncate_tweet(tweet, TWEET_MAX_LEN_WITH_URL)
            tweet = '%s %s' % (tweet, url)
        return truncate_tweet(tweet)

    def _schedule_tweet(self, tweet, profile):
          if not profile:
                self.logger.warning('Empty profile, skipping tweet.')
                return
          self.logger.info('Scheduling tweet on @%s: %s' % (profile.get('twitter_name'), repr(tweet)))
          consumer_key, consumer_secret = self.get_credentials()
          oauth_token = profile.get(KEY_OAUTH_TOKEN)
          oauth_token_secret = profile.get(KEY_OAUTH_TOKEN_SECRET)
          tasks.send_tweet.delay(consumer_key, consumer_secret, oauth_token, oauth_token_secret, tweet)

    def get_vars(self, event):
        settings = SiteSettings.get()

        username = settings.guest_name
        if event.user:
            username = event.user.username

        url = ''
        if event.drink:
            url = event.drink.ShortUrl()
        elif event.session:
            url = event.session.ShortUrl()

        beer_name = ''
        if event.drink and (event.drink.keg and event.drink.keg.type):
            beer_name = event.drink.keg.type.name

        volume_str = ''
        if event.drink:
            volume_str = settings.format_volume(event.drink.volume_ml)

        kbvars = {
            'site_name': settings.title,
            'username': username,
            'url': url,
            'volume_str': volume_str,
            'beer_name': beer_name,
        }
        return kbvars

    def expand_template(self, template, kbvars):
        """Compiles a template from `kbvars` variables."""
        if not template:
            return ''
        template = template.replace('$DRINKER', kbvars.get('username', ''))
        template = template.replace('$VOLUME', kbvars.get('volume_str', ''))
        template = template.replace('$BEER', kbvars.get('beer_name', ''))
        template = template.replace('$SITENAME', kbvars.get('site_name', ''))
        template = template.replace('$URL', kbvars.get('url', ''))
        return template



# Copyright 2011 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Celery tasks for the Kegbot core."""

from kegbot.util import util
from kegbot.util import kbjson

from pykeg.core import models
from pykeg.proto import protolib
from pykeg.plugin import util as plugin_util

from urllib import urlencode
import urllib2

from celery.decorators import task

def schedule_tasks(event_list):
  """Synchronously schedules tasks related to the given events, if any."""
  web_hook_urls = models.SiteSettings.get().web_hook_urls
  if web_hook_urls:
    urls = [u for u in web_hook_urls.split() if u]
    for url in urls:
      post_webhook_event.delay(url, event_list)

  for event in event_list:
    for plugin in plugin_util.get_plugins():
        plugin.handle_new_event(event)

@task
def post_webhook_event(hook_url, event_list):
  post_data = kbjson.dumps({'events': [protolib.ToDict(e) for e in event_list]})
  post_data = urlencode({'payload': post_data})
  opener = urllib2.build_opener()
  opener.addheaders = [
    ('User-agent', 'Kegbot/%s' % util.get_version('kegbot')),
  ]
  try:
    opener.open(hook_url, data=post_data, timeout=5)
    return True
  except urllib2.URLError:
    return False

@task
def handle_new_picture(picture_id):
  pass  # TODO(mikey): plugin support



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for backend"""

from django.test import TestCase
from pykeg.core import models
from pykeg.core import defaults

from . import backend

TAP_NAME = 'kegboard.flow0'
FAKE_BEER_NAME = 'Testy Beer'
FAKE_BREWER_NAME = 'Unittest Brewery'
FAKE_BEER_STYLE = 'Test-Driven Pale Ale'

### Helper methods

class BaseApiTestCase(TestCase):
    def setUp(self):
        self.backend = backend.KegbotBackend()
        defaults.set_defaults(set_is_setup=True)

    def test_drink_management(self):
        """Test adding drinks."""
        tap = models.KegTap.objects.get(meter_name=TAP_NAME)
        tap.ml_per_tick = (1/2.2)
        tap.save()
        keg = self.backend.StartKeg(TAP_NAME, beer_name=FAKE_BEER_NAME,
                brewer_name=FAKE_BREWER_NAME, style_name=FAKE_BEER_STYLE)
        self.assertIsNotNone(keg)

        self.assertEquals(0, keg.served_volume())
        drink = self.backend.RecordDrink(TAP_NAME, ticks=2200)
        self.assertIsNotNone(drink)
        self.assertEquals(2200, drink.ticks)
        self.assertAlmostEqual(1000.0, drink.volume_ml, places=3)
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertAlmostEqual(1000.0, keg.served_volume(), places=3)
        self.assertEquals('', drink.shout)
        self.assertEquals(0, drink.duration)

        drink = self.backend.RecordDrink(TAP_NAME, ticks=1100)
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertIsNotNone(drink)
        self.assertEquals(1100, drink.ticks)
        self.assertAlmostEqual(500.0, drink.volume_ml, places=3)
        self.assertAlmostEqual(1500.0, keg.served_volume(), places=3)

        # Add with volume, taking precedence over ticks.
        drink = self.backend.RecordDrink(TAP_NAME, ticks=1100, volume_ml=1)
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertIsNotNone(drink)
        self.assertEquals(1100, drink.ticks)
        self.assertAlmostEqual(1.0, drink.volume_ml, places=3)
        self.assertAlmostEqual(1501.0, keg.served_volume(), places=3)

        # Add with a user.
        user = models.User.objects.create(username='testy')
        drink = self.backend.RecordDrink(TAP_NAME, ticks=2200, username=user.username)
        self.assertIsNotNone(drink)
        self.assertEquals(user, drink.user)

    def test_drink_cancel(self):
        """Tests cancelling drinks."""
        tap = models.KegTap.objects.get(meter_name=TAP_NAME)
        keg = self.backend.StartKeg(TAP_NAME, beer_name=FAKE_BEER_NAME,
                brewer_name=FAKE_BREWER_NAME, style_name=FAKE_BEER_STYLE)
        self.assertIsNotNone(keg)
        self.assertEquals(0, keg.served_volume())

        for i in xrange(10):
            d = self.backend.RecordDrink(tap_name=TAP_NAME, ticks=1, volume_ml=100)

        drinks = list(models.Drink.objects.all().order_by('id'))
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertEquals(10, len(drinks))
        self.assertAlmostEqual(1000.0, keg.served_volume(), places=3)

        cancel_drink = drinks[-1]
        session = cancel_drink.session
        self.assertAlmostEqual(session.GetStats().total_volume_ml, 1000.0, places=3)

        self.backend.CancelDrink(drinks[-1])
        drinks = list(models.Drink.objects.all().order_by('id'))
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertEquals(9, len(drinks))
        self.assertAlmostEqual(900.0, keg.served_volume(), places=3)
        session = models.DrinkingSession.objects.get(id=session.id)
        self.assertAlmostEqual(session.GetStats().total_volume_ml, 900.0, places=3)

        spilled_drink = drinks[-1]
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertEquals(0, keg.spilled_ml)

        self.backend.CancelDrink(drinks[-1], spilled=True)
        keg = models.Keg.objects.get(pk=keg.id)
        drinks = list(models.Drink.objects.all().order_by('id'))
        self.assertEquals(8, len(drinks))
        self.assertAlmostEqual(800.0, keg.served_volume(), places=3)
        self.assertAlmostEqual(100.0, keg.spilled_ml, places=3)

    def test_keg_management(self):
        """Tests adding and removing kegs."""
        tap = models.KegTap.objects.get(meter_name=TAP_NAME)
        self.assertFalse(tap.is_active(), "Tap is already active.")

        # No beer types yet.
        qs = models.BeerType.objects.filter(name=FAKE_BEER_NAME)
        self.assertEquals(len(qs), 0, "BeerType already exists")

        # Tap the keg.
        keg = self.backend.StartKeg(TAP_NAME, beer_name=FAKE_BEER_NAME,
                brewer_name=FAKE_BREWER_NAME, style_name=FAKE_BEER_STYLE)
        self.assertIsNotNone(keg)
        self.assertTrue(keg.online)

        self.assertEquals(keg.current_tap, tap, "Tap did not become active.")
        tap = keg.current_tap
        self.assertTrue(keg.current_tap.is_active())

        # Check that the beer type was created.
        qs = models.BeerType.objects.filter(name=FAKE_BEER_NAME)
        self.assertEquals(len(qs), 1, "Expected a single new BeerType.")
        beer_type = qs[0]
        brewer = beer_type.brewer
        style = beer_type.style
        self.assertEquals(brewer.name, FAKE_BREWER_NAME)
        self.assertEquals(style.name, FAKE_BEER_STYLE)

        # Now activate a new keg.
        keg_id = keg.id
        keg = self.backend.EndKeg(tap)
        tap = models.KegTap.objects.get(meter_name=TAP_NAME)
        self.assertFalse(tap.is_active())
        self.assertFalse(keg.online)

        new_keg = self.backend.StartKeg(TAP_NAME, beer_type=beer_type)
        self.assertIsNotNone(new_keg)
        self.assertNotEquals(new_keg, keg)

        # Ensure the beer type was reused.
        self.assertEquals(new_keg.type, beer_type)

        # Deactivate, and activate a new keg again by name.
        self.backend.EndKeg(tap.meter_name)
        new_keg_2 = self.backend.StartKeg(TAP_NAME, beer_name='Other Beer',
                brewer_name=FAKE_BREWER_NAME, style_name=FAKE_BEER_STYLE)
        self.assertEquals(new_keg_2.type.brewer, keg.type.brewer)
        self.assertEquals(new_keg_2.type.style, keg.type.style)
        self.assertNotEquals(new_keg_2.type, keg.type)

        # New brewer, identical beer name == new beer type.
        self.backend.EndKeg(tap.meter_name)
        new_keg_3 = self.backend.StartKeg(TAP_NAME, beer_name=FAKE_BEER_NAME,
                brewer_name='Other Brewer', style_name=FAKE_BEER_STYLE)
        self.assertNotEquals(new_keg_3.type.brewer, keg.type.brewer)
        self.assertEquals(new_keg_3.type.name, keg.type.name)
        self.assertNotEquals(new_keg_3.type, keg.type)



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Celery tasks for Foursquare."""

import foursquare
from celery.task import task
import datetime
from pykeg.plugin import util

MAX_CHECKIN_AGE = datetime.timedelta(minutes=30)

logger = util.get_logger(__name__)

@task(expires=60)
def checkin(token, venue_id):
    client = foursquare.Foursquare(access_token=token)
    logger.info('Checkin request: token=%s venue=%s' % (token, venue_id))

    last_checkin = _get_last_checkin(client)
    if last_checkin:
        venue = last_checkin.get('venue', {}).get('id', '')
        if venue == venue_id:
            when = datetime.datetime.fromtimestamp(int(last_checkin.get('createdAt', 0)))
            if (when + MAX_CHECKIN_AGE) > datetime.datetime.now():
                logger.info('Reusing last checkin.')
                return last_checkin
    logger.info('Performing fresh checkin.')
    return client.checkins.add(params={'venueId': venue_id}).get('checkin')


def _get_last_checkin(client):
    result = client.users.checkins(params={'limit': 1})
    items = result.get('checkins', {}).get('items', [])
    if items:
       return items[0]
    return None



from django.http import Http404

from registration.backends import default
from registration.backends import simple

_DEFAULT = default.DefaultBackend()
_SIMPLE = simple.SimpleBackend()

def get_real_backend(request):
  settings = request.kbsite.settings
  if not settings.registration_allowed:
    raise Http404('Registration is disabled.')
  if settings.registration_confirmation:
    return _DEFAULT
  else:
    return _SIMPLE

class KegbotRegistrationBackend:
  """Kegbot-aware backend which delegates to default backends."""
  def register(self, request, **kwargs):
    return get_real_backend(request).register(request, **kwargs)

  def activate(self, request, activation_key):
    return get_real_backend(request).activate(request, **kwargs)

  def registration_allowed(self, request):
    return request.kbsite.settings.registration_allowed

  def get_form_class(self, request):
    return get_real_backend(request).get_form_class(request)

  def post_registration_redirect(self, request, user):
    return ('/account/', (), {})

  def post_activation_redirect(self, request, user):
    return get_real_backend(request).post_activation_redirect(request, user)




#
# Copyright 2008 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb main views."""

import datetime

from django.conf import settings
from django.core.urlresolvers import reverse
from django.contrib.auth.decorators import login_required
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.shortcuts import render_to_response
from django.shortcuts import redirect
from django.template import RequestContext
from django.views.decorators.cache import cache_page
from django.views.generic.dates import ArchiveIndexView
from django.views.generic.dates import DateDetailView
from django.views.generic.dates import DayArchiveView
from django.views.generic.dates import MonthArchiveView
from django.views.generic.dates import YearArchiveView
from django.views.generic.list import ListView

from kegbot.util import kbjson

from pykeg.core import models
from pykeg.proto import protolib

from pykeg.web.kegweb import forms
from pykeg.web.kegweb import signals

### main views

@cache_page(30)
def index(request):
  context = RequestContext(request)

  context['taps'] = models.KegTap.objects.all()

  events = models.SystemEvent.objects.all()[:10]
  context['initial_events'] = kbjson.dumps([protolib.ToDict(e, full=True) for e in events],
      indent=None)

  sessions = models.DrinkingSession.objects.all().order_by('-id')[:10]
  context['sessions'] = sessions
  context['initial_sessions'] = kbjson.dumps([protolib.ToDict(s, full=True) for s in sessions],
      indent=None)

  if sessions:
    last_session = sessions[0]
    context['last_session'] = last_session
    if sessions and last_session.IsActiveNow():
      context['current_session'] = last_session
      if sessions.count() > 1:
        context['last_session'] = sessions[1]
      else:
        context['last_session'] = None

  taps = models.KegTap.objects.filter(current_keg__isnull=False)
  context['initial_taps'] = kbjson.dumps([protolib.ToDict(t, full=True) for t in taps], indent=None)

  context['have_events'] = len(events) > 0
  context['have_taps'] = len(taps) > 0

  return render_to_response('index.html', context_instance=context)

@cache_page(30)
def system_stats(request):
  stats = models.KegbotSite.get().GetStats()
  context = RequestContext(request, {
    'stats': stats,
  })

  top_drinkers = []
  for username, vol in stats.get('volume_by_drinker', {}).iteritems():
    try:
      user = models.User.objects.get(username=username)
    except models.User.DoesNotExist:
      continue  # should not happen
    top_drinkers.append((vol, user))
  top_drinkers.sort(reverse=True)

  context['top_drinkers'] = top_drinkers[:10]

  return render_to_response('kegweb/system-stats.html', context_instance=context)


### object lists and detail (generic views)

def user_detail(request, username):
  user = get_object_or_404(models.User, username=username, is_active=True)
  stats = user.get_profile().GetStats()

  # TODO(mikey): Limit to "recent" sessions for now.
  # Bug: https://github.com/Kegbot/kegbot/issues/37
  chunks = models.UserSessionChunk.objects.filter(user=user).select_related(depth=1)[:10]
  drinks = user.drinks.all()

  context = RequestContext(request, {
      'drinks': drinks,
      'chunks': chunks,
      'stats': stats,
      'drinker': user})
  return render_to_response('kegweb/drinker_detail.html', context_instance=context)

class KegListView(ListView):
  model = models.Keg
  template_name = 'kegweb/keg_list.html'
  context_object_name = 'kegs'
  paginate_by = 10

  def get_queryset(self):
    return models.Keg.objects.all().order_by('-id')

@cache_page(30)
def keg_detail(request, keg_id):
  keg = get_object_or_404(models.Keg, id=keg_id)
  sessions = keg.Sessions()
  context = RequestContext(request, {
    'keg': keg,
    'stats': keg.GetStats(),
    'sessions': sessions})
  return render_to_response('kegweb/keg_detail.html', context_instance=context)

def short_drink_detail(request, drink_id):
  return redirect('kb-drink', drink_id=str(drink_id), permanent=True)

def short_session_detail(request, session_id):
  session = get_object_or_404(models.DrinkingSession, id=session_id)
  url = session.get_absolute_url()
  return redirect(url, permanent=True)

def drink_detail(request, drink_id):
  drink = get_object_or_404(models.Drink, id=drink_id)
  context = RequestContext(request, {'drink': drink})
  return render_to_response('kegweb/drink_detail.html', context_instance=context)

def session_detail(request, year, month, day, id, slug):
  session = get_object_or_404(models.DrinkingSession, id=id)
  context = RequestContext(request, {
    'session': session,
    'stats': session.GetStats(),
  })
  return render_to_response('kegweb/session_detail.html', context_instance=context)


class SessionArchiveIndexView(ArchiveIndexView):
  model = models.DrinkingSession
  date_field = 'start_time'
  template_name = 'kegweb/drinkingsession_archive.html'
  context_object_name = 'sessions'
  paginate_by = 20


class SessionYearArchiveView(YearArchiveView):
  model = models.DrinkingSession
  date_field = 'start_time'
  template_name = 'kegweb/drinkingsession_archive_year.html'
  make_object_list = True
  context_object_name = 'sessions'
  paginate_by = 20


class SessionMonthArchiveView(MonthArchiveView):
  model = models.DrinkingSession
  date_field = 'start_time'
  template_name = 'kegweb/drinkingsession_archive_month.html'
  make_object_list = True
  context_object_name = 'sessions'
  paginate_by = 20


class SessionDayArchiveView(DayArchiveView):
  model = models.DrinkingSession
  date_field = 'start_time'
  template_name = 'kegweb/drinkingsession_archive_day.html'
  make_object_list = True
  context_object_name = 'sessions'
  paginate_by = 20


class SessionDateDetailView(DateDetailView):
  model = models.DrinkingSession
  date_field = 'start_time'
  template_name = 'kegweb/session_detail.html'
  context_object_name = 'session'

  def get_context_data(self, **kwargs):
    """Adds `stats` to the context."""
    ret = super(SessionDateDetailView, self).get_context_data(**kwargs)
    ret['stats'] = ret[self.context_object_name].GetStats()
    return ret




# -*- coding: latin-1 -*-
# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import pytz
import random

from django.conf import settings
from django.core.urlresolvers import reverse
from django.db import models
from django.db.models.signals import post_save
from django.db.models.signals import pre_save
from django.contrib.auth.models import User
from django.utils import timezone

from pykeg import EPOCH

from pykeg.core import kb_common
from pykeg.core import fields
from pykeg.core import imagespecs
from pykeg.core import jsonfield
from pykeg.core import managers
from pykeg.core.util import make_serial

from kegbot.util import units
from kegbot.util import util

from kegbot.api import models_pb2
from kegbot.api import protoutil

"""Django models definition for the kegbot database."""

TIMEZONE_CHOICES = ((z, z) for z in pytz.common_timezones)

class KegbotSite(models.Model):
  name = models.CharField(max_length=64, unique=True, default='default',
      editable=False)
  is_active = models.BooleanField(default=True,
      help_text='On/off switch for this site.')
  is_setup = models.BooleanField(default=False,
      help_text='True if the site has completed setup.',
      editable=False)
  epoch = models.PositiveIntegerField(default=EPOCH,
      help_text='Database epoch number.',
      editable=False)
  serial_number = models.TextField(max_length=128, editable=False,
      blank=True, default='',
      help_text='A unique id for this system.')

  def __str__(self):
    return self.name

  @classmethod
  def get(cls):
    """Gets the default site settings."""
    return KegbotSite.objects.get_or_create(name='default',
        defaults={'is_setup': False})[0]

  def GetStatsRecord(self):
    try:
      return SystemStats.objects.latest()
    except SystemStats.DoesNotExist:
      return None

  def GetStats(self):
    ret = {}
    record = self.GetStatsRecord()
    if record:
      ret = record.stats
    return util.AttrDict(ret)

def _kegbotsite_pre_save(sender, instance, **kwargs):
  if not instance.serial_number:
    instance.serial_number = make_serial()
pre_save.connect(_kegbotsite_pre_save, sender=KegbotSite)

def _kegbotsite_post_save(sender, instance, **kwargs):
  """Creates a SiteSettings object if none already exists."""
  settings, _ = SiteSettings.objects.get_or_create(site=instance)
post_save.connect(_kegbotsite_post_save, sender=KegbotSite)

class SiteSettings(models.Model):
  """General system-wide settings."""
  VOLUME_DISPLAY_UNITS_CHOICES = (
    ('metric', 'Metric (mL, L)'),
    ('imperial', 'Imperial (oz, pint)'),
  )
  TEMPERATURE_DISPLAY_UNITS_CHOICES = (
    ('f', 'Fahrenheit'),
    ('c', 'Celsius'),
  )
  PRIVACY_CHOICES = (
    ('public', 'Public: Browsing does not require login'),
    ('members', 'Members only: Must log in to browse'),
    ('staff', 'Staff only: Only logged-in staff accounts may browse'),
  )
  DEFAULT_PRIVACY = 'public'

  site = models.OneToOneField(KegbotSite, related_name='settings')
  volume_display_units = models.CharField(max_length=64,
      choices=VOLUME_DISPLAY_UNITS_CHOICES, default='imperial',
      help_text='Unit system to use when displaying volumetric data.')
  temperature_display_units = models.CharField(max_length=64,
      choices=TEMPERATURE_DISPLAY_UNITS_CHOICES, default='f',
      help_text='Unit system to use when displaying temperature data.')
  title = models.CharField(max_length=64, default='My Kegbot',
      help_text='The title of this site.')
  description = models.TextField(blank=True, null=True,
      help_text='Description of this site')
  background_image = models.ForeignKey('Picture', blank=True, null=True,
      on_delete=models.SET_NULL,
      help_text='Background for this site.')
  web_hook_urls = models.TextField(blank=True, null=True,
      help_text='URL(s) to POST newly-generated events.')
  google_analytics_id = models.CharField(blank=True, null=True, max_length=64,
      help_text='Set to your Google Analytics ID to enable tracking. '
      'Example: UA-XXXX-y')
  session_timeout_minutes = models.PositiveIntegerField(
      default=kb_common.DRINK_SESSION_TIME_MINUTES,
      help_text='Maximum time, in minutes, that a session may be idle (no pours) '
          'before it is considered to be finished.  '
          'Recommended value is %s.' % kb_common.DRINK_SESSION_TIME_MINUTES)
  privacy = models.CharField(max_length=63, choices=PRIVACY_CHOICES,
      default=DEFAULT_PRIVACY,
      help_text='Whole-system setting for system privacy.')
  guest_name = models.CharField(max_length=63, default='guest',
      help_text='Name to be shown in various places for unauthenticated pours.')
  guest_image = models.ForeignKey('Picture', blank=True, null=True,
      related_name='guest_images', on_delete=models.SET_NULL,
      help_text='Profile picture to be shown for unauthenticated pours.')
  default_user = models.ForeignKey(User, blank=True, null=True,
      help_text='Default user to set as owner for unauthenticated drinks. '
          'When set, the "guest" user will not be used. This is mostly '
          'useful for closed, single-user systems.')
  registration_allowed = models.BooleanField(default=True,
      help_text='Whether to allow new user registration.')
  registration_confirmation = models.BooleanField(default=False,
      help_text='Whether registration requires e-mail confirmation.')
  allowed_hosts = models.TextField(blank=True, null=True, default='',
      help_text='List of allowed hostnames. If blank, validation is disabled.')
  timezone = models.CharField(max_length=255, choices=TIMEZONE_CHOICES,
      default='UTC',
      help_text='Time zone for this system')
  hostname = models.CharField(max_length=255,
      help_text='Hostname (and optional port) for this system. Examples: mykegbot.example.com, 192.168.1.100:8000')
  use_ssl = models.BooleanField(default=False,
      help_text='Use SSL for URLs to this site.')

  class Meta:
    verbose_name_plural = "site settings"

  def GetSessionTimeoutDelta(self):
    return datetime.timedelta(minutes=self.session_timeout_minutes)

  def base_url(self):
    protocol = 'http'
    if self.use_ssl:
      protocol = 'https'
    return '%s://%s' % (protocol, self.hostname)

  def reverse_full(self, *args, **kwargs):
    """Calls reverse, and returns a full URL (includes base_url())."""
    return '%s%s' % (self.base_url(), reverse(*args, **kwargs))

  def format_volume(self, volume_ml):
    if SiteSettings.get().volume_display_units == 'metric':
      if volume_ml < 500:
        return '%d mL' % int(volume_ml)
      return '%.1f L' % (volume_ml / 1000.0)
    else:
      return '%1.f oz' % units.Quantity(volume_ml).InOunces()

  @classmethod
  def get(cls):
    """Gets the default site settings."""
    return KegbotSite.get().settings


class UserProfile(models.Model):
  """Extra per-User information."""
  def __str__(self):
    return "profile for %s" % (self.user,)

  def FacebookProfile(self):
    if 'socialregistration' not in settings.INSTALLED_APPS:
      return None
    qs = self.user.facebookprofile_set.all()
    if qs:
      return qs[0]

  def TwitterProfile(self):
    if 'socialregistration' not in settings.INSTALLED_APPS:
      return None
    qs = self.user.twitterprofile_set.all()
    if qs:
      return qs[0]

  def GetStatsRecord(self):
    qs = UserStats.objects.filter(user=self.user).order_by('-id')
    if len(qs):
      return qs[0]
    return None

  def GetStats(self):
    ret = {}
    record = self.GetStatsRecord()
    if record:
      ret = record.stats
    return util.AttrDict(ret)

  def GetApiKey(self):
    api_key, new = ApiKey.objects.get_or_create(user=self.user,
        defaults={'key': ApiKey.generate_key()})
    return api_key.key

  def get_absolute_url(self):
    return reverse('kb-drinker', args=(self.user.username,))

  user = models.OneToOneField(User)
  mugshot = models.ForeignKey('Picture', blank=True, null=True,
    on_delete=models.SET_NULL)

def _user_post_save(sender, instance, **kwargs):
  profile, new = UserProfile.objects.get_or_create(user=instance)
post_save.connect(_user_post_save, sender=User)


class ApiKey(models.Model):
  """Grants access to certain API endpoints to a user via a secret key."""
  user = models.OneToOneField(User,
      help_text='User receiving API access.')
  key = models.CharField(max_length=127, editable=False, unique=True,
      help_text='The secret key.')
  active = models.BooleanField(default=True,
      help_text='Whether access by this key is currently allowed.')

  def is_active(self):
    """Returns true if both the key and the key's user are active."""
    return self.active and self.user.is_active

  def regenerate(self):
    """Discards and regenerates the key."""
    self.key = self.generate_key()
    self.save()

  @classmethod
  def generate_key(cls):
    """Returns a new random key."""
    return '%032x' % random.randint(0, 2**128 - 1)


class BeerDBModel(models.Model):
  class Meta:
    abstract = True
  added = models.DateTimeField(default=timezone.now, editable=False)
  edited = models.DateTimeField(editable=False)
  beerdb_id = models.CharField(blank=True, null=True, max_length=128)

  def save(self, *args, **kwargs):
    self.edited = timezone.now()
    super(BeerDBModel, self).save(*args, **kwargs)


class Brewer(BeerDBModel):
  """Describes a producer of beer."""
  class Meta:
        ordering = ('name',)

  PRODUCTION_CHOICES = (
    ('commercial', 'Commercial brewer'),
    ('homebrew', 'Home brewer'),
  )

  name = models.CharField(max_length=255,
      help_text='Name of the brewer')
  country = fields.CountryField(default='USA',
      help_text='Country of origin')
  origin_state = models.CharField(max_length=128,
      default='', blank=True, null=True,
      help_text='State of origin, if applicable')
  origin_city = models.CharField(max_length=128, default='', blank=True,
      null=True,
      help_text='City of origin, if known')
  production = models.CharField(max_length=128, choices=PRODUCTION_CHOICES,
      default='commercial')
  url = models.URLField(default='', blank=True, null=True,
      help_text='Brewer\'s home page')
  description = models.TextField(default='', blank=True, null=True,
      help_text='A short description of the brewer')
  image = models.ForeignKey('Picture', blank=True, null=True,
      related_name='beer_brewers', on_delete=models.SET_NULL)

  def __str__(self):
    return self.name


class BeerStyle(BeerDBModel):
  """Describes a named style of beer (Stout, IPA, etc)"""
  class Meta:
        ordering = ('name',)

  name = models.CharField(max_length=128,
      help_text='Name of the beer style')

  def __str__(self):
    return self.name


class BeerType(BeerDBModel):
  """Describes a specific kind of beer, by name, brewer, and style."""
  class Meta:
        ordering = ('name',)

  name = models.CharField(max_length=255,
      help_text='Name of the beer; typically unique within a Brewer.')
  brewer = models.ForeignKey(Brewer,
      help_text='Brewer producing this beer.')
  style = models.ForeignKey(BeerStyle,
      help_text='Style of the beer.')
  edition = models.CharField(max_length=255, blank=True, null=True,
      help_text='For seasonal or special edition beers, the year or '
          'other name uniquely identifying it.')
  abv = models.FloatField(blank=True, null=True,
      help_text='Alcohol by volume, as percentage (0-100).')
  calories_oz = models.FloatField(blank=True, null=True,
      help_text='Calories per fluid ounce of beer.')
  carbs_oz = models.FloatField(blank=True, null=True,
      help_text='Carbohydrates per fluid ounce of beer.')
  original_gravity = models.FloatField(blank=True, null=True,
      help_text='Original gravity of the beer.')
  specific_gravity = models.FloatField(blank=True, null=True,
      help_text='Specific/final gravity of the beer.')
  image = models.ForeignKey('Picture', blank=True, null=True,
      related_name='beer_types', on_delete=models.SET_NULL,
      help_text='Logo or artwork for this beer type.')
  untappd_beer_id = models.IntegerField(blank=True, null=True,
      help_text='Untappd.com beer id for this beer, if known')

  def __str__(self):
    return "%s by %s" % (self.name, self.brewer)

  def GetImage(self):
    if self.image:
      return self.image
    if self.brewer.image:
      return self.brewer.image
    return None


class KegSize(models.Model):
  """ A convenient table of common Keg sizes """
  name = models.CharField(max_length=128,
      help_text='Name of the size, for example "Half Barrel".')
  volume_ml = models.FloatField(help_text='Volume per instance.')

  def __str__(self):
    gallons = units.Quantity(self.volume_ml).InUSGallons()
    return "%s [%.2f gal]" % (self.name, gallons)


class KegTap(models.Model):
  """A physical tap of beer."""
  meter_name = models.CharField(max_length=128, unique=True,
      help_text='Flow meter name for this tap, for example, "kegboard.flow0"')
  name = models.CharField(max_length=128,
      help_text='The display name for this tap, for example, "Main Tap".')
  relay_name = models.CharField(max_length=128, blank=True, null=True,
      help_text='Name of relay attached to this tap.')
  ml_per_tick = models.FloatField(default=kb_common.ML_PER_TICK,
      help_text='mL per flow meter tick.  Common values: %s '
      '(SwissFlow), 0.454545454545 (Vision 2000)' % kb_common.ML_PER_TICK)
  description = models.TextField(blank=True, null=True,
      help_text='User-visible description for this tap.')
  current_keg = models.OneToOneField('Keg', blank=True, null=True,
      related_name='current_tap',
      help_text='Keg currently connected to this tap.')
  temperature_sensor = models.ForeignKey('ThermoSensor', blank=True, null=True,
      on_delete=models.SET_NULL,
      help_text='Sensor monitoring the temperature of this Keg.')

  def __str__(self):
    return "%s: %s" % (self.meter_name, self.name)

  def is_active(self):
    """Returns True if the tap has an active Keg."""
    return self.current_keg is not None

  def Temperature(self):
    if self.temperature_sensor:
      last_rec = self.temperature_sensor.thermolog_set.all().order_by('-time')
      if last_rec:
        return last_rec[0]
    return None


class Keg(models.Model):
  """Record for each physical Keg."""
  type = models.ForeignKey(BeerType, on_delete=models.PROTECT,
      help_text='Beverage in this Keg.')
  size = models.ForeignKey(KegSize, on_delete=models.PROTECT,
      help_text='Size of this Keg.')
  served_volume_ml = models.FloatField(default=0, editable=False,
      help_text='Computed served volume.')
  full_volume_ml = models.FloatField(default=0, editable=False,
      help_text='Full volume of this Keg, usually set from its KegSize.')
  start_time = models.DateTimeField(default=timezone.now,
      help_text='Time the Keg was first tapped.')
  end_time = models.DateTimeField(default=timezone.now,
      help_text='Time the Keg was finished or disconnected.')
  online = models.BooleanField(default=True, editable=False,
      help_text='True if the keg is currently assigned to a tap.')
  finished = models.BooleanField(default=True, editable=False,
      help_text='True when the Keg has been exhausted or discarded.')
  description = models.CharField(max_length=256, blank=True, null=True,
      help_text='User-visible description of the Keg.')
  spilled_ml = models.FloatField(default=0, editable=False,
      help_text='Amount of beverage poured without an associated Drink.')
  notes = models.TextField(blank=True, null=True,
      help_text='Private notes about this keg, viewable only by admins.')

  def get_absolute_url(self):
    return reverse('kb-keg', args=(str(self.id),))

  def full_volume(self):
    return self.full_volume_ml

  def served_volume(self):
    # Deprecated
    return self.served_volume_ml

  def remaining_volume(self):
    return self.full_volume_ml - self.served_volume_ml

  def percent_full(self):
    result = float(self.remaining_volume()) / float(self.full_volume_ml) * 100
    result = max(min(result, 100), 0)
    return result

  def keg_age(self):
    if self.online:
      end = timezone.now()
    else:
      end = self.end_time
    return end - self.start_time

  def is_empty(self):
    return float(self.remaining_volume()) <= 0

  def previous(self):
    q = Keg.objects.filter(start_time__lt=self.start_time).order_by('-start_time')
    if q.count():
      return q[0]
    return None

  def next(self):
    q = Keg.objects.filter(start_time__gt=self.start_time).order_by('start_time')
    if q.count():
      return q[0]
    return None

  def GetStatsRecord(self):
    qs = KegStats.objects.filter(keg=self).order_by('-id')
    if len(qs):
      return qs[0]
    return None

  def GetStats(self):
    ret = {}
    record = self.GetStatsRecord()
    if record:
      ret = record.stats
    return util.AttrDict(ret)

  def Sessions(self):
    chunks = SessionChunk.objects.filter(keg=self).order_by('-start_time').select_related(depth=2)
    sessions = []
    sess = None
    for c in chunks:
      # Skip same sessions
      if c.session == sess:
        continue
      sess = c.session
      sessions.append(sess)
    return sessions

  def TopDrinkers(self):
    stats = self.GetStats()
    if not stats:
      return []
    ret = []
    entries = stats.get('volume_by_drinker', {})
    for username, vol in entries.iteritems():
      try:
        user = User.objects.get(username=username)
      except User.DoesNotExist:
        continue  # should not happen
      ret.append((vol, user))
    ret.sort(reverse=True)
    return ret

  def __str__(self):
    return "Keg #%s - %s" % (self.id, self.type)

def _keg_pre_save(sender, instance, **kwargs):
  keg = instance
  # We don't need to do anything if the keg is still online.
  if keg.online:
    return

  # Determine first drink date & set keg start date to it if earlier.
  drinks = keg.drinks.all().order_by('time')
  if drinks:
    drink = drinks[0]
    if drink.time < keg.start_time:
      keg.start_time = drink.time

  # Determine last drink date & set keg end date to it if later.
  drinks = keg.drinks.all().order_by('-time')
  if drinks:
    drink = drinks[0]
    if drink.time > keg.end_time:
      keg.end_time = drink.time

pre_save.connect(_keg_pre_save, sender=Keg)

def _keg_post_save(sender, instance, **kwargs):
  keg = instance
  SystemEvent.ProcessKeg(keg)

post_save.connect(_keg_post_save, sender=Keg)


class Drink(models.Model):
  """ Table of drinks records """
  class Meta:
    get_latest_by = 'time'
    ordering = ('-time',)

  ticks = models.PositiveIntegerField(editable=False,
      help_text='Flow sensor ticks, never changed once recorded.')
  volume_ml = models.FloatField(editable=False,
      help_text='Calculated (or set) Drink volume.')
  time = models.DateTimeField(editable=False,
    help_text='Date and time of pour.')
  duration = models.PositiveIntegerField(blank=True, default=0, editable=False,
      help_text='Time in seconds taken to pour this Drink.')
  user = models.ForeignKey(User, null=True, blank=True, related_name='drinks',
      editable=False,
      help_text='User responsible for this Drink, or None if anonymous/unknown.')
  keg = models.ForeignKey(Keg, null=True, blank=True, related_name='drinks',
      on_delete=models.PROTECT, editable=False,
      help_text='Keg against which this Drink is accounted.')
  session = models.ForeignKey('DrinkingSession',
      related_name='drinks', null=True, blank=True, editable=False,
      on_delete=models.PROTECT,
      help_text='Session where this Drink is grouped.')
  shout = models.TextField(blank=True, null=True,
      help_text='Comment from the drinker at the time of the pour.')
  tick_time_series = models.TextField(blank=True, null=True, editable=False,
      help_text='Tick update sequence that generated this drink (diagnostic data).')

  def get_absolute_url(self):
    return reverse('kb-drink', args=(str(self.id),))

  def ShortUrl(self):
    return '%s%s' % (SiteSettings.get().base_url(), reverse('kb-drink-short', args=(str(self.id),)))

  def Volume(self):
    return units.Quantity(self.volume_ml)

  def calories(self):
    if not self.keg or not self.keg.type:
      return 0
    ounces = self.Volume().InOunces()
    return self.keg.type.calories_oz * ounces

  def __str__(self):
    return "Drink %i by %s" % (self.id, self.user)


class AuthenticationToken(models.Model):
  """A secret token to authenticate a user, optionally pin-protected."""
  class Meta:
    unique_together = ('auth_device', 'token_value')

  auth_device = models.CharField(max_length=64,
      help_text='Namespace for this token.')
  token_value = models.CharField(max_length=128,
      help_text='Actual value of the token, unique within an auth_device.')
  nice_name = models.CharField(max_length=256, blank=True, null=True,
      help_text='A human-readable alias for the token, for example "Guest Key".')
  pin = models.CharField(max_length=256, blank=True, null=True,
      help_text='A secret value necessary to authenticate with this token.')
  user = models.ForeignKey(User, blank=True, null=True,
      related_name='tokens',
      help_text='User in possession of and authenticated by this token.')
  enabled = models.BooleanField(default=True,
      help_text='Whether this token is considered active.')
  created_time = models.DateTimeField(auto_now_add=True,
      help_text='Date token was first added to the system.')
  expire_time = models.DateTimeField(blank=True, null=True,
      help_text='Date after which token is treated as disabled.')

  def __str__(self):
    auth_device = self.auth_device
    if auth_device == 'core.rfid':
      auth_device = 'RFID'
    elif auth_device == 'core.onewire':
      auth_device = 'OneWire'

    ret = "%s %s" % (auth_device, self.token_value)
    if self.nice_name:
      ret += " (%s)" % self.nice_name
    return ret

  def get_auth_device(self):
    auth_device = self.auth_device
    if auth_device == 'core.rfid':
      auth_device = 'RFID'
    elif auth_device == 'core.onewire':
      auth_device = 'OneWire'
    return auth_device

  def IsAssigned(self):
    return self.user is not None

  def IsActive(self):
    if not self.enabled:
      return False
    if not self.expire_time:
      return True
    return timezone.now() < self.expire_time

def _auth_token_pre_save(sender, instance, **kwargs):
  if instance.auth_device in kb_common.AUTH_MODULE_NAMES_HEX_VALUES:
    instance.token_value = instance.token_value.lower()

pre_save.connect(_auth_token_pre_save, sender=AuthenticationToken)

class _AbstractChunk(models.Model):
  class Meta:
    abstract = True
    get_latest_by = 'start_time'
    ordering = ('-start_time',)

  start_time = models.DateTimeField()
  end_time = models.DateTimeField()
  volume_ml = models.FloatField(default=0)

  def Duration(self):
    return self.end_time - self.start_time

  def _AddDrinkNoSave(self, drink):
    session_delta = SiteSettings.get().GetSessionTimeoutDelta()
    session_end = drink.time + session_delta

    if self.start_time > drink.time:
      self.start_time = drink.time
    if self.end_time < session_end:
      self.end_time = session_end
    self.volume_ml += drink.volume_ml

  def AddDrink(self, drink):
    self._AddDrinkNoSave(drink)
    self.save()


class DrinkingSession(_AbstractChunk):
  """A collection of contiguous drinks. """
  class Meta:
    get_latest_by = 'start_time'
    ordering = ('-start_time',)

  objects = managers.SessionManager()
  name = models.CharField(max_length=256, blank=True, null=True)

  def __str__(self):
    return "Session #%s: %s" % (self.id, self.start_time)

  def ShortUrl(self):
    return '%s%s' % (SiteSettings.get().base_url(), reverse('kb-session-short',
        args=(str(self.id),)))

  def HighlightPicture(self):
    pictures = self.pictures.all().order_by('-time')
    if pictures:
      return pictures[0]
    chunks = self.user_chunks.filter(user__ne=None).order_by('-volume_ml')
    if chunks:
      mugshot = chunks[0].user.get_profile().mugshot
      return mugshot

  def OtherPictures(self):
    pictures = self.pictures.all().order_by('-time')
    if pictures:
      return pictures[1:]
    return []

  def get_absolute_url(self):
    dt = self.start_time
    if settings.USE_TZ:
      dt = timezone.localtime(dt)
    return reverse('kb-session-detail', args=(), kwargs={
      'year' : dt.year,
      'month' : dt.month,
      'day' : dt.day,
      'pk' : self.pk})

  def GetStatsRecord(self):
    qs = SessionStats.objects.filter(session=self).order_by('-id')
    if len(qs):
      return qs[0]
    return None

  def GetStats(self):
    ret = {}
    record = self.GetStatsRecord()
    if record:
      ret = record.stats
    return util.AttrDict(ret)

  def summarize_drinkers(self):
    def fmt(user):
      url = '/drinkers/%s/' % (user.username,)
      return '<a href="%s">%s</a>' % (url, user.username)
    chunks = self.user_chunks.all().order_by('-volume_ml')
    users = tuple(c.user for c in chunks)
    names = tuple(fmt(u) for u in users if u)

    if None in users:
      guest_trailer = ' (and possibly others)'
    else:
      guest_trailer = ''

    num = len(names)
    if num == 0:
      return 'no known drinkers'
    elif num == 1:
      ret = names[0]
    elif num == 2:
      ret = '%s and %s' % names
    elif num == 3:
      ret = '%s, %s and %s' % names
    else:
      if guest_trailer:
        return '%s, %s and at least %i others' % (names[0], names[1], num-2)
      else:
        return '%s, %s and %i others' % (names[0], names[1], num-2)

    return '%s%s' % (ret, guest_trailer)

  def GetTitle(self):
    if self.name:
      return self.name
    else:
      if self.id:
        return 'Session %s' % (self.id,)
      else:
        # Not yet saved.
        return 'New Session'

  def AddDrink(self, drink):
    super(DrinkingSession, self).AddDrink(drink)
    session_delta = SiteSettings.get().GetSessionTimeoutDelta()

    defaults = {
      'start_time': drink.time,
      'end_time': drink.time + session_delta,
    }

    # Update or create a SessionChunk.
    chunk, created = SessionChunk.objects.get_or_create(session=self,
        user=drink.user, keg=drink.keg, defaults=defaults)
    chunk.AddDrink(drink)

    # Update or create a UserSessionChunk.
    chunk, created = UserSessionChunk.objects.get_or_create(session=self,
        user=drink.user, defaults=defaults)
    chunk.AddDrink(drink)

    # Update or create a KegSessionChunk.
    chunk, created = KegSessionChunk.objects.get_or_create(session=self,
        keg=drink.keg, defaults=defaults)
    chunk.AddDrink(drink)

  def UserChunksByVolume(self):
    chunks = self.user_chunks.all().order_by('-volume_ml')
    return chunks

  def IsActiveNow(self):
    return self.IsActive(timezone.now())

  def IsActive(self, now):
    return self.end_time > now

  def Rebuild(self):
    """Recomputes start and end time, and chunks, based on current drinks.

    This method should be called after changing the set of drinks
    belonging to this session.

    This method has no effect on statistics; see stats module.
    """
    self.volume_ml = 0
    self.chunks.all().delete()
    self.user_chunks.all().delete()
    self.keg_chunks.all().delete()

    drinks = self.drinks.all()
    if not drinks:
      self.save()
      # TODO(mikey): cancel/delete the session entirely.  As it is, session will
      # remain a placeholder.
      return

    session_delta = SiteSettings.get().GetSessionTimeoutDelta()
    min_time = None
    max_time = None
    for d in drinks:
      self.AddDrink(d)
      if min_time is None or d.time < min_time:
        min_time = d.time
      if max_time is None or d.time > max_time:
        max_time = d.time
    self.start_time = min_time
    self.end_time = max_time + session_delta
    self.save()

  @classmethod
  def AssignSessionForDrink(cls, drink):
    # Return existing session if already assigned.
    if drink.session:
      return drink.session

    # Return last session if one already exists
    q = DrinkingSession.objects.all().order_by('-end_time')[:1]
    if q and q[0].IsActive(drink.time):
      session = q[0]
      session.AddDrink(drink)
      drink.session = session
      drink.save()
      return session

    # Create a new session
    session = cls(start_time=drink.time, end_time=drink.time)
    session.save()
    session.AddDrink(drink)
    drink.session = session
    drink.save()
    return session


class SessionChunk(_AbstractChunk):
  """A specific user and keg contribution to a session."""
  class Meta:
    unique_together = ('session', 'user', 'keg')
    get_latest_by = 'start_time'
    ordering = ('-start_time',)

  session = models.ForeignKey(DrinkingSession, related_name='chunks')
  user = models.ForeignKey(User, related_name='session_chunks', blank=True,
      null=True)
  keg = models.ForeignKey(Keg, related_name='session_chunks', blank=True,
      null=True, on_delete=models.PROTECT)


class UserSessionChunk(_AbstractChunk):
  """A specific user's contribution to a session (spans all kegs)."""
  class Meta:
    unique_together = ('session', 'user')
    get_latest_by = 'start_time'
    ordering = ('-start_time',)

  session = models.ForeignKey(DrinkingSession, related_name='user_chunks')
  user = models.ForeignKey(User, related_name='user_session_chunks', blank=True,
      null=True)

  def GetTitle(self):
    return self.session.GetTitle()

  def GetDrinks(self):
    return self.session.drinks.filter(user=self.user).order_by('time')


class KegSessionChunk(_AbstractChunk):
  """A specific keg's contribution to a session (spans all users)."""
  class Meta:
    unique_together = ('session', 'keg')
    get_latest_by = 'start_time'
    ordering = ('-start_time',)

  objects = managers.SessionManager()
  session = models.ForeignKey(DrinkingSession, related_name='keg_chunks')
  keg = models.ForeignKey(Keg, related_name='keg_session_chunks', blank=True,
      null=True)

  def GetTitle(self):
    return self.session.GetTitle()


class ThermoSensor(models.Model):
  raw_name = models.CharField(max_length=256)
  nice_name = models.CharField(max_length=128)

  def __str__(self):
    if self.nice_name:
      return '%s (%s)' % (self.nice_name, self.raw_name)
    return self.raw_name

  def LastLog(self):
    try:
      return self.thermolog_set.latest()
    except Thermolog.DoesNotExist:
      return None


class Thermolog(models.Model):
  """ A log from an ITemperatureSensor device of periodic measurements. """
  class Meta:
    get_latest_by = 'time'
    ordering = ('-time',)

  sensor = models.ForeignKey(ThermoSensor)
  temp = models.FloatField()
  time = models.DateTimeField()

  def __str__(self):
    return '%s %.2f C / %.2f F [%s]' % (self.sensor, self.TempC(),
        self.TempF(), self.time)

  def TempC(self):
    return self.temp

  def TempF(self):
    return util.CtoF(self.temp)


class _StatsModel(models.Model):
  time = models.DateTimeField(default=timezone.now)
  stats = jsonfield.JSONField()
  drink = models.ForeignKey(Drink)

  class Meta:
    abstract = True
    get_latest_by = 'id'


class SystemStats(_StatsModel):
  pass

class UserStats(_StatsModel):
  user = models.ForeignKey(User, blank=True, null=True,
      related_name='stats')

  class Meta:
    unique_together = ('drink', 'user')

  def __str__(self):
    return 'UserStats for %s' % self.user


class KegStats(_StatsModel):
  keg = models.ForeignKey(Keg, related_name='stats')
  completed = models.BooleanField(default=False)

  class Meta:
    unique_together = ('drink', 'keg')

  def __str__(self):
    return 'KegStats for %s' % self.keg


class SessionStats(_StatsModel):
  session = models.ForeignKey(DrinkingSession, related_name='stats')
  completed = models.BooleanField(default=False)

  class Meta:
    unique_together = ('drink', 'session')

  def __str__(self):
    return 'SessionStats for %s' % self.session


class SystemEvent(models.Model):
  class Meta:
    ordering = ('-id',)
    get_latest_by = 'time'

  KINDS = (
      ('drink_poured', 'Drink poured'),
      ('session_started', 'Session started'),
      ('session_joined', 'User joined session'),
      ('keg_tapped', 'Keg tapped'),
      ('keg_ended', 'Keg ended'),
  )

  kind = models.CharField(max_length=255, choices=KINDS,
      help_text='Type of event.')
  time = models.DateTimeField(help_text='Time of the event.')
  user = models.ForeignKey(User, blank=True, null=True,
      related_name='events',
      help_text='User responsible for the event, if any.')
  drink = models.ForeignKey(Drink, blank=True, null=True,
      related_name='events',
      help_text='Drink involved in the event, if any.')
  keg = models.ForeignKey(Keg, blank=True, null=True,
      related_name='events',
      help_text='Keg involved in the event, if any.')
  session = models.ForeignKey(DrinkingSession, blank=True, null=True,
      related_name='events',
      help_text='Session involved in the event, if any.')

  objects = managers.SystemEventManager()

  def __str__(self):
    if self.kind == 'drink_poured':
      ret = 'Drink %i poured' % self.drink.id
    elif self.kind == 'session_started':
      ret = 'Session %s started by drink %s' % (self.session.id,
          self.drink.id)
    elif self.kind == 'session_joined':
      ret = 'Session %s joined by %s (drink %s)' % (self.session.id,
          self.user.username, self.drink.id)
    elif self.kind == 'keg_tapped':
      ret = 'Keg %s tapped' % self.keg.id
    elif self.kind == 'keg_ended':
      ret = 'Keg %s ended' % self.keg.id
    else:
      ret = 'Unknown event type (%s)' % self.kind
    return 'Event %s: %s' % (self.id, ret)

  @classmethod
  def ProcessKeg(cls, keg):
    if keg.online:
      q = keg.events.filter(kind='keg_tapped')
      if q.count() == 0:
        e = keg.events.create(kind='keg_tapped', time=keg.start_time, keg=keg)
        e.save()

    if not keg.online:
      q = keg.events.filter(kind='keg_ended')
      if q.count() == 0:
        e = keg.events.create(kind='keg_ended', time=keg.end_time, keg=keg)
        e.save()

  @classmethod
  def ProcessDrink(cls, drink):
    keg = drink.keg
    session = drink.session
    user = drink.user

    events = []
    if keg:
      q = keg.events.filter(kind='keg_tapped')
      if q.count() == 0:
        e = keg.events.create(kind='keg_tapped', time=drink.time,
            keg=keg, user=user, drink=drink, session=session)
        e.save()
        events.append(e)

    if session:
      q = session.events.filter(kind='session_started')
      if q.count() == 0:
        e = session.events.create(kind='session_started',
            time=session.start_time, drink=drink, user=user)
        e.save()
        events.append(e)

    if user:
      q = user.events.filter(kind='session_joined', session=session)
      if q.count() == 0:
        e = user.events.create(kind='session_joined',
            time=drink.time, session=session, drink=drink, user=user)
        e.save()
        events.append(e)

    q = drink.events.filter(kind='drink_poured')
    if q.count() == 0:
      e = drink.events.create(kind='drink_poured',
          time=drink.time, drink=drink, user=user, keg=keg,
          session=session)
      e.save()
      events.append(e)

    return events


def _pics_file_name(instance, filename):
  rand_salt = random.randrange(0xffff)
  new_filename = '%04x-%s' % (rand_salt, filename)
  return os.path.join('pics', new_filename)

class Picture(models.Model):
  image = models.ImageField(upload_to=_pics_file_name,
      help_text='The image')
  resized = imagespecs.resized
  thumbnail = imagespecs.thumbnail
  small_resized = imagespecs.small_resized
  small_thumbnail = imagespecs.small_thumbnail

  time = models.DateTimeField(default=timezone.now)

  def __str__(self):
    return 'Picture: %s' % self.image


class PourPicture(models.Model):
  '''Stores additional metadata about a picture taken during a pour.'''
  class Meta:
    get_latest_by = 'time'
    ordering = ('-time',)

  picture = models.ForeignKey('Picture')
  drink = models.ForeignKey(Drink, blank=True, null=True,
      related_name='pictures',
      help_text='Drink this picture is associated with, if any')
  time = models.DateTimeField(default=timezone.now)
  caption = models.TextField(blank=True, null=True,
      help_text='Caption for the picture')
  user = models.ForeignKey(User, blank=True, null=True,
      help_text='User this picture is associated with, if any')
  keg = models.ForeignKey(Keg, blank=True, null=True, related_name='pictures',
      on_delete=models.SET_NULL,
      help_text='Keg this picture is associated with, if any')
  session = models.ForeignKey(DrinkingSession, blank=True, null=True,
      on_delete=models.SET_NULL,
      related_name='pictures',
      help_text='Session this picture is associated with, if any')

  def GetCaption(self):
    if self.caption:
      return self.caption
    elif self.drink:
      if self.user:
        return '%s pouring drink %s' % (self.user.username, self.drink.id)
      else:
        return 'An unknown drinker pouring drink %s' % (self.drink.id,)
    else:
      return ''




# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding field 'SiteSettings.hostname'
        db.add_column(u'core_sitesettings', 'hostname',
                      self.gf('django.db.models.fields.CharField')(default='kegbot.example.com', max_length=255),
                      keep_default=False)

        # Adding field 'SiteSettings.use_ssl'
        db.add_column(u'core_sitesettings', 'use_ssl',
                      self.gf('django.db.models.fields.BooleanField')(default=False),
                      keep_default=False)


    def backwards(self, orm):
        # Deleting field 'SiteSettings.hostname'
        db.delete_column(u'core_sitesettings', 'hostname')

        # Deleting field 'SiteSettings.use_ssl'
        db.delete_column(u'core_sitesettings', 'use_ssl')


    models = {
        u'auth.group': {
            'Meta': {'object_name': 'Group'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        u'auth.permission': {
            'Meta': {'ordering': "(u'content_type__app_label', u'content_type__model', u'codename')", 'unique_together': "((u'content_type', u'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        u'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'core.apikey': {
            'Meta': {'object_name': 'ApiKey'},
            'active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '127'}),
            'user': ('django.db.models.fields.related.OneToOneField', [], {'to': u"orm['auth.User']", 'unique': 'True'})
        },
        u'core.authenticationtoken': {
            'Meta': {'unique_together': "(('auth_device', 'token_value'),)", 'object_name': 'AuthenticationToken'},
            'auth_device': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'created_time': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'expire_time': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'nice_name': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'pin': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'token_value': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'tokens'", 'null': 'True', 'to': u"orm['auth.User']"})
        },
        u'core.beerstyle': {
            'Meta': {'ordering': "('name',)", 'object_name': 'BeerStyle'},
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'beerdb_id': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'edited': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'})
        },
        u'core.beertype': {
            'Meta': {'ordering': "('name',)", 'object_name': 'BeerType'},
            'abv': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'beerdb_id': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'brewer': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Brewer']"}),
            'calories_oz': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'carbs_oz': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'edited': ('django.db.models.fields.DateTimeField', [], {}),
            'edition': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'image': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'beer_types'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Picture']"}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'original_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'specific_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'style': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.BeerStyle']"}),
            'untappd_beer_id': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'})
        },
        u'core.brewer': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Brewer'},
            'added': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'beerdb_id': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'country': ('pykeg.core.fields.CountryField', [], {'default': "'USA'", 'max_length': '3'}),
            'description': ('django.db.models.fields.TextField', [], {'default': "''", 'null': 'True', 'blank': 'True'}),
            'edited': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'image': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'beer_brewers'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Picture']"}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'origin_city': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'origin_state': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'production': ('django.db.models.fields.CharField', [], {'default': "'commercial'", 'max_length': '128'}),
            'url': ('django.db.models.fields.URLField', [], {'default': "''", 'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        u'core.drink': {
            'Meta': {'ordering': "('-time',)", 'object_name': 'Drink'},
            'duration': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'drinks'", 'null': 'True', 'on_delete': 'models.PROTECT', 'to': u"orm['core.Keg']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'drinks'", 'null': 'True', 'on_delete': 'models.PROTECT', 'to': u"orm['core.DrinkingSession']"}),
            'shout': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'tick_time_series': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'ticks': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'drinks'", 'null': 'True', 'to': u"orm['auth.User']"}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {})
        },
        u'core.drinkingsession': {
            'Meta': {'ordering': "('-start_time',)", 'object_name': 'DrinkingSession'},
            'end_time': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        u'core.keg': {
            'Meta': {'object_name': 'Keg'},
            'description': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'end_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'finished': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'full_volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'online': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'served_volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'size': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.KegSize']", 'on_delete': 'models.PROTECT'}),
            'spilled_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.BeerType']", 'on_delete': 'models.PROTECT'})
        },
        u'core.kegbotsite': {
            'Meta': {'object_name': 'KegbotSite'},
            'epoch': ('django.db.models.fields.PositiveIntegerField', [], {'default': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_setup': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'default'", 'unique': 'True', 'max_length': '64'}),
            'serial_number': ('django.db.models.fields.TextField', [], {'default': "''", 'max_length': '128', 'blank': 'True'})
        },
        u'core.kegsessionchunk': {
            'Meta': {'ordering': "('-start_time',)", 'unique_together': "(('session', 'keg'),)", 'object_name': 'KegSessionChunk'},
            'end_time': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'keg_session_chunks'", 'null': 'True', 'to': u"orm['core.Keg']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'keg_chunks'", 'to': u"orm['core.DrinkingSession']"}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        u'core.kegsize': {
            'Meta': {'object_name': 'KegSize'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {})
        },
        u'core.kegstats': {
            'Meta': {'unique_together': "(('drink', 'keg'),)", 'object_name': 'KegStats'},
            'completed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'to': u"orm['core.Keg']"}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {'default': "'{}'"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'})
        },
        u'core.kegtap': {
            'Meta': {'object_name': 'KegTap'},
            'current_keg': ('django.db.models.fields.related.OneToOneField', [], {'blank': 'True', 'related_name': "'current_tap'", 'unique': 'True', 'null': 'True', 'to': u"orm['core.Keg']"}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'meter_name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '128'}),
            'ml_per_tick': ('django.db.models.fields.FloatField', [], {'default': '0.45454545454545453'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'relay_name': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'temperature_sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.ThermoSensor']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'})
        },
        u'core.picture': {
            'Meta': {'object_name': 'Picture'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'image': ('django.db.models.fields.files.ImageField', [], {'max_length': '100'}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'})
        },
        u'core.pourpicture': {
            'Meta': {'ordering': "('-time',)", 'object_name': 'PourPicture'},
            'caption': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Keg']"}),
            'picture': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.DrinkingSession']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        u'core.sessionchunk': {
            'Meta': {'ordering': "('-start_time',)", 'unique_together': "(('session', 'user', 'keg'),)", 'object_name': 'SessionChunk'},
            'end_time': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'session_chunks'", 'null': 'True', 'on_delete': 'models.PROTECT', 'to': u"orm['core.Keg']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'chunks'", 'to': u"orm['core.DrinkingSession']"}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'session_chunks'", 'null': 'True', 'to': u"orm['auth.User']"}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        u'core.sessionstats': {
            'Meta': {'unique_together': "(('drink', 'session'),)", 'object_name': 'SessionStats'},
            'completed': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'to': u"orm['core.DrinkingSession']"}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {'default': "'{}'"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'})
        },
        u'core.sitesettings': {
            'Meta': {'object_name': 'SiteSettings'},
            'allowed_hosts': ('django.db.models.fields.TextField', [], {'default': "''", 'null': 'True', 'blank': 'True'}),
            'background_image': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'default_user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['auth.User']", 'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'google_analytics_id': ('django.db.models.fields.CharField', [], {'max_length': '64', 'null': 'True', 'blank': 'True'}),
            'guest_image': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'guest_images'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Picture']"}),
            'guest_name': ('django.db.models.fields.CharField', [], {'default': "'guest'", 'max_length': '63'}),
            'hostname': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'privacy': ('django.db.models.fields.CharField', [], {'default': "'public'", 'max_length': '63'}),
            'registration_allowed': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'registration_confirmation': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'session_timeout_minutes': ('django.db.models.fields.PositiveIntegerField', [], {'default': '180'}),
            'site': ('django.db.models.fields.related.OneToOneField', [], {'related_name': "'settings'", 'unique': 'True', 'to': u"orm['core.KegbotSite']"}),
            'temperature_display_units': ('django.db.models.fields.CharField', [], {'default': "'f'", 'max_length': '64'}),
            'timezone': ('django.db.models.fields.CharField', [], {'default': "'UTC'", 'max_length': '255'}),
            'title': ('django.db.models.fields.CharField', [], {'default': "'My Kegbot'", 'max_length': '64'}),
            'use_ssl': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'volume_display_units': ('django.db.models.fields.CharField', [], {'default': "'imperial'", 'max_length': '64'}),
            'web_hook_urls': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'})
        },
        u'core.systemevent': {
            'Meta': {'ordering': "('-id',)", 'object_name': 'SystemEvent'},
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.Keg']"}),
            'kind': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.DrinkingSession']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['auth.User']"})
        },
        u'core.systemstats': {
            'Meta': {'object_name': 'SystemStats'},
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {'default': "'{}'"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'})
        },
        u'core.thermolog': {
            'Meta': {'ordering': "('-time',)", 'object_name': 'Thermolog'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.ThermoSensor']"}),
            'temp': ('django.db.models.fields.FloatField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {})
        },
        u'core.thermosensor': {
            'Meta': {'object_name': 'ThermoSensor'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'nice_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'raw_name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'core.userprofile': {
            'Meta': {'object_name': 'UserProfile'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'mugshot': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'user': ('django.db.models.fields.related.OneToOneField', [], {'to': u"orm['auth.User']", 'unique': 'True'})
        },
        u'core.usersessionchunk': {
            'Meta': {'ordering': "('-start_time',)", 'unique_together': "(('session', 'user'),)", 'object_name': 'UserSessionChunk'},
            'end_time': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'user_chunks'", 'to': u"orm['core.DrinkingSession']"}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'user_session_chunks'", 'null': 'True', 'to': u"orm['auth.User']"}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        u'core.userstats': {
            'Meta': {'unique_together': "(('drink', 'user'),)", 'object_name': 'UserStats'},
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {'default': "'{}'"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'stats'", 'null': 'True', 'to': u"orm['auth.User']"})
        }
    }

    complete_apps = ['core']


# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import messages
from django.core.urlresolvers import reverse
from django.shortcuts import redirect
from socialregistration.clients.oauth import OAuthError
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import render_to_response
from django.template import RequestContext
import oauth2 as oauth

from . import forms
from . import oauth_client

@staff_member_required
def admin_settings(request, plugin):
  context = RequestContext(request)
  settings_form = plugin.get_site_settings_form()

  if request.method == 'POST':
    if 'submit-settings' in request.POST:
      settings_form = forms.SiteSettingsForm(request.POST)
      if settings_form.is_valid():
        client_id = settings_form.cleaned_data['client_id']
        client_secret = settings_form.cleaned_data['client_secret']
        plugin.save_site_settings_form(settings_form)
        messages.success(request, 'Settings updated')

  context['plugin'] = plugin
  context['settings_form'] = settings_form

  return render_to_response('contrib/untappd/untappd_admin_settings.html',
      context_instance=context)


@login_required
def user_settings(request, plugin):
  context = RequestContext(request)
  user = request.user

  settings_form = plugin.get_user_settings_form(user)

  if request.method == 'POST':
    if 'submit-settings' in request.POST:
      settings_form = forms.UserSettingsForm(request.POST)
      if settings_form.is_valid():
        plugin.save_user_settings_form(user, settings_form)
        messages.success(request, 'Settings updated')

  context['plugin'] = plugin
  context['profile'] = plugin.get_user_profile(user)
  context['settings_form'] = settings_form

  return render_to_response('contrib/untappd/untappd_user_settings.html',
      context_instance=context)


@staff_member_required
def auth_redirect(request):
  if 'submit-remove' in request.POST:
    plugin = request.plugins.get('untappd')
    plugin.save_user_profile(request.user, None)
    plugin.save_user_token(request.user, '')
    messages.success(request, 'Removed Untappd account.')
    return redirect('account-plugin-settings', plugin_name='untappd')

  plugin = request.plugins['untappd']
  url = request.kbsite.settings.reverse_full('plugin-untappd-callback')
  client = get_client(*plugin.get_credentials(), callback_url=url)

  request.session['untappd_client'] = client

  try:
    return redirect(client.get_redirect_url())
  except OAuthError, error:
    messages.error(request, 'Error: %s' % str(error))
    return redirect('account-plugin-settings', plugin_name='untappd')


@staff_member_required
def auth_callback(request):
  try:
    client = request.session['untappd_client']
    del request.session['untappd_client']
    token = client.complete(dict(request.GET.items()))
  except KeyError:
    messages.error(request, 'Session expired.')
  except OAuthError, error:
    messages.error(request, str(error))
  else:
    plugin = request.plugins.get('untappd')
    profile = client.get_user_info()
    token = client.get_access_token()
    plugin.save_user_profile(request.user, profile)
    plugin.save_user_token(request.user, token)

    username = '%s %s' % (profile.get('first_name'), profile.get('last_name'))
    messages.success(request, 'Successfully linked to Untappd user %s' % username)

  return redirect('account-plugin-settings', plugin_name='untappd')


def get_client(client_id, client_secret, callback_url):
  client = oauth_client.Untappd(callback_url=callback_url)
  client.client_id = client_id
  client.secret = client_secret
  return client


# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.core.management.base import NoArgsCommand

from pykeg.core import backend
from pykeg.core import models
from pykeg.core.management.commands.common import progbar

DRINKERS = (
  'mike', 'brian', 'dom', 'eric', 'ted', 'justin', None
)
NUM_DRINKS = 100
TAP = 'kegboard.flow0'

import time
import random
import datetime
from django.utils import timezone
BASE_TIME = timezone.now() - datetime.timedelta(days=365)

#import logging
#l = logging.getLogger('django.db.backends')
#l.setLevel(logging.DEBUG)
#l.addHandler(logging.StreamHandler())

class Command(NoArgsCommand):
  help = u'Generate fake data'
  args = '<none>'

  def handle(self, **options):
    profile = False
    if profile:
      import cProfile
      command = 'self.generate()'
      cProfile.runctx(command, globals(), locals())
    else:
      self.generate()

  def generate(self):
    num_drinks = 0
    num_sessions = 0

    now = BASE_TIME
    b = backend.KegbotBackend()
    while num_drinks < NUM_DRINKS:
      num_drinks += 1
      ticks = 400 + random.randrange(0, 400)
      username = random.choice(DRINKERS)
      if username:
        models.User.objects.get_or_create(username=username)

      start_time = time.time()
      d = b.RecordDrink(TAP, ticks, pour_time=now, username=username)
      delta = time.time() - start_time
      end_session = random.random() >= 0.75

      ms = int(delta * 1000)
      print '%s: %s: %s [%s]' % (ms, d.id, d.user, d.keg)

      if end_session:
        now += datetime.timedelta(hours=24 + random.randrange(0, 12))
      else:
        now += datetime.timedelta(minutes=random.randrange(3, 60))

      if d.keg.remaining_volume < 10000:
        b.EndKeg(TAP)
        b.StartKeg(TAP)




# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for pykeg.web.api"""

from django.test import TestCase
from pykeg.core import models
from pykeg.core import defaults
from kegbot.util import kbjson

### Helper methods

def create_site():
    return defaults.set_defaults(set_is_setup=True)

class BaseApiTestCase(TestCase):
    def get(self, subpath, data={}, follow=False, **extra):
        response = self.client.get('/api/%s' % subpath, data=data, follow=follow,
            **extra)
        return response, kbjson.loads(response.content)

class ApiClientTestCase(BaseApiTestCase):
    def testNotSetUp(self):
        '''Api endpoints should all error out prior to site setup.'''

        endpoints = ('events/', 'taps/', 'kegs/')
        for endpoint in endpoints:
            response, data = self.get(endpoint)
            self.assertEquals(data.meta.result, 'error')
            self.assertEquals(data.error.code, 'BadRequestError')

        create_site()

        # Ordinary results expected after site installed.
        for endpoint in endpoints:
            response, data = self.get(endpoint)
            self.assertEquals(data.meta.result, 'ok')

    def testSiteDefaults(self):
        create_site()

        empty_endpoints = ('events/', 'kegs/')
        for endpoint in empty_endpoints:
            response, data = self.get(endpoint)
            self.assertEquals(data.objects, [])

        response, data = self.get('taps/')
        taps = data.objects
        self.assertEquals(2, len(taps))
        self.assertEquals('Main Tap', taps[0].name)
        self.assertEquals('kegboard.flow0', taps[0].meter_name)
        self.assertEquals('Second Tap', taps[1].name)
        self.assertEquals('kegboard.flow1', taps[1].meter_name)

    def testApiAccess(self):
        site = create_site()

        protected_get_endpoints = ('users/',)

        user = models.User.objects.create(username='testuser')
        user.set_password('testpass')
        user.save()

        endpoint = 'users/'

        # No API key.
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'NoAuthTokenError')

        # Non-existent key.
        response, data = self.get(endpoint, HTTP_X_KEGBOT_API_KEY='123')
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'BadApiKeyError')

        api_key_obj = models.ApiKey.objects.create(user=user, key='123')

        # Key exists, but non-superuser.
        response, data = self.get(endpoint, HTTP_X_KEGBOT_API_KEY='123')
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'PermissionDeniedError')

        user.is_staff = True
        user.save()

        # Finally ok.
        response, data = self.get(endpoint, HTTP_X_KEGBOT_API_KEY='123')
        self.assertEquals(data.meta.result, 'ok')

        endpoint = 'events/'

        # Alter privacy and compare.
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'ok')
        site.settings.privacy = 'members'
        site.settings.save()

        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'error')

        self.client.login(username='testuser', password='testpass')
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'ok')

        # Alert to staff-only.
        site.settings.privacy = 'staff'
        site.settings.save()
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'ok')

        user.is_staff = False
        user.save()
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'NoAuthTokenError')







from django.conf.urls import include
from django.conf.urls import patterns
from django.conf.urls import url

urlpatterns = patterns('pykeg.web.kegadmin.views',
    ### main page
    url(r'^$', 'dashboard', name='kegadmin-dashboard'),
    url(r'^settings/$', 'general_settings', name='kegadmin-main'),

    url(r'^beers/$', 'beer_type_list', name='kegadmin-beer-types'),
    url(r'^beers/(?P<beer_id>\d+)/$', 'beer_type_detail', name='kegadmin-edit-beer-type'),

    url(r'^beer-styles/$', 'beer_style_list', name='kegadmin-beer-styles'),
    url(r'^beer-styles/(?P<style_id>\d+)/$', 'beer_style_detail', name='kegadmin-edit-beer-style'),

    url(r'^brewers/$', 'brewer_list', name='kegadmin-brewers'),
    url(r'^brewers/(?P<brewer_id>\d+)/$', 'brewer_detail', name='kegadmin-edit-brewer'),

    url(r'^taps/$', 'tap_list', name='kegadmin-taps'),
    url(r'^taps/create/$', 'add_tap', name='kegadmin-add-tap'),
    url(r'^taps/(?P<tap_id>\d+)/$', 'tap_detail', name='kegadmin-edit-tap'),

    url(r'^users/$', 'user_list', name='kegadmin-users'),
    url(r'^users/create/$', 'add_user', name='kegadmin-add-user'),
    url(r'^users/(?P<user_id>\d+)/$', 'user_detail', name='kegadmin-edit-user'),

    url(r'^drinks/$', 'drink_list', name='kegadmin-drinks'),
    url(r'^drinks/(?P<drink_id>\d+)/$', 'drink_edit', name='kegadmin-edit-drink'),

    url(r'^tokens/$', 'token_list', name='kegadmin-tokens'),
    url(r'^tokens/create/$', 'add_token', name='kegadmin-add-token'),
    url(r'^tokens/(?P<token_id>\d+)/$', 'token_detail', name='kegadmin-edit-token'),

    url(r'^backup-restore/$', 'backup_restore', name='kegadmin-backup-restore'),
    url(r'^backup-restore/dump/$', 'generate_backup', name='kegadmin-get-backup'),
    url(r'^logs/$', 'logs', name='kegadmin-logs'),
    url(r'^autocomplete/beer/$', 'autocomplete_beer_type',
      name='kegadmin-autocomplete-beer'),
    url(r'^autocomplete/user/$', 'autocomplete_user',
      name='kegadmin-autocomplete-user'),
    url(r'^autocomplete/token/$', 'autocomplete_token',
      name='kegadmin-autocomplete-token'),

    url(r'^plugin/(?P<plugin_name>\w+)/$', 'plugin_settings', name='kegadmin-plugin-settings'),
)

from pykeg.plugin import util
if util.get_plugins():
    urlpatterns += util.get_admin_urls()



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Foursquare plugin forms."""

from django import forms
from django.conf import settings

EMBEDDED = getattr(settings, 'EMBEDDED', False)
WIDE_TEXT = forms.TextInput(attrs={'class': 'input-block-level'})

class SiteSettingsForm(forms.Form):
    if not EMBEDDED:
        client_id = forms.CharField(required=False, widget=WIDE_TEXT,
            help_text='Foursquare API Client ID.')
        client_secret = forms.CharField(required=False, widget=WIDE_TEXT,
            help_text='Foursquare API Client Secret')
    venue_id = forms.CharField(required=False, widget=WIDE_TEXT,
        help_text='Venue ID for this Kegbot; will be used for checkins.')

class UserSettingsForm(forms.Form):
    enable_checkins = forms.BooleanField(initial=True, required=False,
        help_text='Check in when you join a session.')
    attach_photos = forms.BooleanField(initial=True, required=False,
        help_text='Attach photos as you drink.')



# Copyright 2003-2012 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import logging
import traceback
import datetime

from collections import deque

PREFIX = 'pykeg.core.logger:'
CACHED_LOGS_KEY = PREFIX + 'buffer'

class CacheHandler(logging.Handler):
  """An exception log handler that buffers errors in the Django cache for future
  processing.
  """
  def __init__(self, max_entries=10):
    logging.Handler.__init__(self)
    self.max_entries = max_entries

  def emit(self, record):
    """Adds the record to the cache."""
    from django.core.cache import cache
    cached_entries = get_cached_logs()
    if cached_entries is not None:
      entries = cached_entries
    else:
      entries = deque()

    # Oldest entries on the left.
    while len(entries) >= self.max_entries:
      entries.pop()

    entries.appendleft(format_record(record))
    cache.set(CACHED_LOGS_KEY, entries)

def format_record(record):
  from django.conf import settings
  ret = {}

  ret['when'] = datetime.datetime.now()
  ret['level'] = record.levelname
  ret['summary'] = record.getMessage()

  if hasattr(record, 'request'):
    request = record.request
    ret['addr'] = request.META.get('REMOTE_ADDR')
    ret['request_path'] = request.path
    ret['method'] = request.method
    if hasattr(request, 'user'):
      ret['username'] = request.user.username
    else:
      ret['usernmae'] = ''
  else:
    ret['addr'] = ''
    ret['request_path'] = ''
    ret['method'] = ''
    ret['username'] = ''

  if record.exc_info:
    ret['exc_text'] = '\n'.join(traceback.format_exception(*record.exc_info))
  else:
    ret['exc_text'] = ''

  return ret

class RequireDebugTrue(logging.Filter):
  def filter(self, record):
    from django.conf import settings
    return settings.DEBUG

def get_cached_logs():
  from django.core.cache import cache
  return cache.get(CACHED_LOGS_KEY)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'PluginData'
        db.create_table(u'plugin_plugindata', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('plugin_name', self.gf('django.db.models.fields.CharField')(max_length=127)),
            ('key', self.gf('django.db.models.fields.CharField')(max_length=127)),
            ('value', self.gf('django.db.models.fields.TextField')()),
        ))
        db.send_create_signal(u'plugin', ['PluginData'])

        # Adding unique constraint on 'PluginData', fields ['plugin_name', 'key']
        db.create_unique(u'plugin_plugindata', ['plugin_name', 'key'])


    def backwards(self, orm):
        # Removing unique constraint on 'PluginData', fields ['plugin_name', 'key']
        db.delete_unique(u'plugin_plugindata', ['plugin_name', 'key'])

        # Deleting model 'PluginData'
        db.delete_table(u'plugin_plugindata')


    models = {
        u'plugin.plugindata': {
            'Meta': {'unique_together': "(('plugin_name', 'key'),)", 'object_name': 'PluginData'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'max_length': '127'}),
            'plugin_name': ('django.db.models.fields.CharField', [], {'max_length': '127'}),
            'value': ('django.db.models.fields.TextField', [], {})
        }
    }

    complete_apps = ['plugin']


# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Celery tasks for Untappd."""

import requests
import urllib
import datetime
from pykeg.plugin import util
from celery.task import task

logger = util.get_logger(__name__)

@task(expires=60)
def checkin(token, beer_id, timezone_name, shout=None):
    logger.info('Checking in: token=%s beer_id=%s timezone_name=%s' % (token, beer_id, timezone_name))

    # TODO(mikey): API does not appear to support Olso tz names for
    # the timezone parameter.  Report as GMT, as it doesn't seem to
    # be meaningful.
    data = {
        'bid': beer_id,
        'gmt_offset': 0,
        'timezone': 'GMT',
    }

    if shout:
        data['shout'] = shout

    params = {'access_token': token}
    url = 'https://api.untappd.com/v4/checkin/add'
    r = requests.post(url, params=params, data=data)
    logger.info('Response: %s' % str(r.text))





# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from functools import wraps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate
from django.contrib.auth import login
from django.core.urlresolvers import reverse
from django.http import Http404
from django.shortcuts import redirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.views.decorators.cache import never_cache

from pykeg.core import defaults

from .forms import AdminUserForm
from .forms import CreateOrImportForm
from .forms import MiniSiteSettingsForm

def setup_view(f):
  """Decorator for setup views."""
  def new_function(*args, **kwargs):
    request = args[0]
    if not settings.DEBUG:
      raise Http404('Site is not in DEBUG mode.')
    if request.kbsite and request.kbsite.is_setup:
      raise Http404('Site is already setup, wizard disabled.')
    return f(*args, **kwargs)
  return wraps(f)(new_function)

@setup_view
@never_cache
def start(request):
  context = RequestContext(request)
  return render_to_response('setup_wizard/start.html', context_instance=context)

@setup_view
@never_cache
def create_or_import(request):
  context = RequestContext(request)
  form = CreateOrImportForm(initial={'mode': 'create'})
  if request.method == 'POST':
    form = CreateOrImportForm(request.POST)
    if form.is_valid():
      if form.cleaned_data['mode'] == 'create':
        try:
          defaults.set_defaults()
          messages.success(request, 'Started new site!')
        except defaults.AlreadyInstalledError:
          messages.warning(request, 'Site already installed, proceeding.')
        return redirect('setup_site_settings')
      else:
        messages.error(request, 'Sorry, imports are not yet supported.')
  context['form'] = form
  print ' '.join(str(type(d)) for d in context.dicts)
  return render_to_response('setup_wizard/create_or_import.html', context_instance=context)

@setup_view
@never_cache
def site_settings(request):
  context = RequestContext(request)
  form = MiniSiteSettingsForm(instance=request.kbsite.settings)
  if request.method == 'POST':
    form = MiniSiteSettingsForm(request.POST, instance=request.kbsite.settings)
    if form.is_valid():
      form.save()
      messages.success(request, 'Settings saved!')
      return redirect('setup_admin')
  context['form'] = form
  return render_to_response('setup_wizard/site_settings.html', context_instance=context)

@setup_view
@never_cache
def admin(request):
  context = RequestContext(request)
  form = AdminUserForm()
  if request.method == 'POST':
    form = AdminUserForm(request.POST)
    if form.is_valid():
      form.save()
      user = authenticate(username=form.cleaned_data.get('username'),
          password=form.cleaned_data.get('password'))
      if user:
        login(request, user)
      return redirect('setup_finish')
  context['form'] = form
  return render_to_response('setup_wizard/admin.html', context_instance=context)

@setup_view
@never_cache
def finish(request):
  context = RequestContext(request)
  if request.method == 'POST':
    request.kbsite.is_setup = True
    request.kbsite.save()
    messages.success(request, 'Tip: Install a new Keg in Admin: Taps')
    return redirect('kegadmin-main')
  return render_to_response('setup_wizard/finish.html', context_instance=context)



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for setup wizard."""

from django.conf import settings
from django.core.cache import cache
from django.test import TransactionTestCase
from django.test.utils import override_settings

from . import models
from pykeg.core import defaults

class SetupWizardTestCase(TransactionTestCase):

    def setUp(self):
        cache.clear()

    def test_settings_debug_false(self):
        """Verify wizard is not offered (DEBUG is False)."""
        for path in ('/', '/stats/'):
            response = self.client.get(path)
            self.assertContains(response, '<h2>Kegbot Offline</h2>',
                status_code=403)
            self.assertNotContains(response, 'Start Setup',
                status_code=403)

        response = self.client.get('/setup/')
        self.failUnlessEqual(response.status_code, 404)


    @override_settings(DEBUG=True)
    def test_settings_debug_true(self):
        """Verify wizard is offered (DEBUG is True)."""
        for path in ('/', '/stats/'):
            response = self.client.get(path)
            self.assertContains(response, '<h2>Setup Required</h2>', status_code=403)
            self.assertContains(response, 'Start Setup', status_code=403)

        response = self.client.get('/setup/')
        self.failUnlessEqual(response.status_code, 200)


    def test_setup_not_shown(self):
        """Verify wizard is not shown on set-up site."""
        site = defaults.set_defaults()
        site.is_setup = True
        site.save()
        for path in ('/', '/stats/'):
            response = self.client.get(path)
            self.assertNotContains(response, '<h2>Kegbot Offline</h2>',
                status_code=200)
            self.assertNotContains(response, '<h2>Setup Required</h2>',
                status_code=200)
            self.assertNotContains(response, 'Start Setup', status_code=200)

        response = self.client.get('/setup/')
        self.failUnlessEqual(response.status_code, 404)



# Copyright 2003-2011 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.test import TransactionTestCase

from kegbot.api import models_pb2
from kegbot.api.protoutil import ProtoMessageToDict
from kegbot.util import util

from . import backend
from . import models
from . import stats
from .testutils import make_datetime

class StatsTestCase(TransactionTestCase):
  reset_sequences = True

  def setUp(self):
    self.backend = backend.KegbotBackend()

    test_usernames = ('user1', 'user2', 'user3')
    self.users = [self.backend.CreateNewUser(name) for name in test_usernames]

    self.taps = [
        self.backend.CreateTap('tap1', 'kegboard.flow0', ml_per_tick=1/2200.0),
        self.backend.CreateTap('tap2', 'kegboard.flow1', ml_per_tick=1/2200.0),
    ]

    self.keg = self.backend.StartKeg('kegboard.flow0', beer_name='Unknown',
        brewer_name='Unknown', style_name='Unknown')

  def testStuff(self):
    site = models.KegbotSite.get()
    stats = site.GetStats()
    self.assertEquals(stats, {})

    now = make_datetime(2012, 1, 2, 12, 00)
    self.maxDiff = None

    d = self.backend.RecordDrink('kegboard.flow0', ticks=1, volume_ml=100,
        username='user1', pour_time=now)
    expected = util.AttrDict({
        u'volume_by_year': {u'2012': 100.0},
        u'total_pours': 1,
        u'has_guest_pour': False,
        u'greatest_volume_ml': 100.0,
        u'registered_drinkers': [u'user1'],
        u'volume_by_day_of_week': {u'1': 100.0},
        u'greatest_volume_id': d.id,
        u'volume_by_drinker': {u'user1': 100.0},
        u'last_drink_id': d.id,
        u'sessions_count': 1,
        u'average_volume_ml': 100.0,
        u'total_volume_ml': 100.0
    })
    stats = site.GetStats()
    self.assertDictEqual(expected, stats)

    now = make_datetime(2012, 1, 3, 12, 00)
    d = self.backend.RecordDrink('kegboard.flow0', ticks=200,
        volume_ml=200, username='user2', pour_time=now)
    stats = site.GetStats()
    expected.total_pours = 2
    expected.greatest_volume_ml = 200.0
    expected.greatest_volume_id = d.id
    expected.registered_drinkers.append(u'user2')
    expected.volume_by_drinker[u'user2'] = 200.0
    expected.last_drink_id = d.id
    expected.average_volume_ml = 150.0
    expected.total_volume_ml = 300.0
    expected.volume_by_day_of_week[u'2'] = 200.0
    expected.volume_by_year[u'2012'] = 300.0
    expected.sessions_count = 2
    print 'EXPECTED'
    import pprint
    pprint.pprint(expected)
    print '----'
    print 'ACTUAL'
    pprint.pprint(stats)
    self.assertDictEqual(expected, stats)





from django.conf import settings
from django.contrib.sites.models import Site
from django.core.urlresolvers import reverse
from socialregistration.clients.oauth import OAuth2
from socialregistration.settings import SESSION_KEY
import json
import urllib


class Untappd(OAuth2):
    client_id = getattr(settings, 'UNTAPPD_CLIENT_ID', '')
    secret = getattr(settings, 'UNTAPPD_CLIENT_SECRET', '')
    auth_url = 'https://untappd.com/oauth/authenticate'
    access_token_url = 'https://untappd.com/oauth/authorize/'

    def __init__(self, *args, **kwargs):
        self._user_info = None
        self.callback_url = kwargs.pop('callback_url')
        super(Untappd, self).__init__(*args, **kwargs)

    def get_callback_url(self, **kwargs):
        return self.callback_url

    def get_redirect_url(self, state='', **kwargs):
        params = {
            'response_type': 'code',
            'client_id': self.client_id,
            'redirect_url': self.get_callback_url(**kwargs),
            'state': state,
        }

        return '%s?%s' % (self.auth_url, urllib.urlencode(params))

    def parse_access_token(self, content):
        """
        Untappd returns JSON instead of url encoded data.
        """
        response = json.loads(content)
        print 'HERE!', response
        if not response or not response.get('response'):
            raise ValueError, "Malformed response: %s" % str(response)

        return response.get('response', {})

    def request_access_token(self, params):
        params['redirect_url'] = params['redirect_uri']
        del params['redirect_uri']
        params['response_type'] = 'code'
        url = '%s?%s' % (self.access_token_url, urllib.urlencode(params))
        print 'URL:', url
        return self.request(url, method="GET")

    def get_user_info(self):
        if self._user_info is None:
            resp, content = self.request('https://api.untappd.com/v4/user/info')
            self._user_info = json.loads(content)['response']['user']
        return self._user_info

    @staticmethod
    def get_session_key():
        return '%suntappd' % SESSION_KEY



# Copyright 2013 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for Twitter plugin."""

import datetime

from unittest import TestCase
from pykeg.contrib.twitter import plugin

TAP_NAME = 'kegboard.flow0'

class FakeDatastore:
    def __init__(self):
        self.backing_dict = dict()

    def set(self, key, value):
        self.backing_dict[key] = value

    def get(self, key, default=None):
        return self.backing_dict.get(key, default)

    def delete(self, key):
        del self.backing_dict[key]

class TwitterPluginTestCase(TestCase):
    def setUp(self):
        self.datastore = FakeDatastore()
        self.backing_dict = self.datastore.backing_dict
        self.plugin = plugin.TwitterPlugin(self.datastore)

    def testTruncateTweet(self):
        msg = 'x'*138 + 'x'
        truncated = plugin.truncate_tweet(msg)
        self.assertEquals(msg, truncated)

        longer_msg = 'x'*138 + ' xxxx'
        truncated = plugin.truncate_tweet(longer_msg)
        self.assertEqual(139, len(truncated))
        self.assertEqual('x'*138 + plugin.TRUNCATE_STR, truncated)

    def testPluginBasics(self):
        self.assertEquals({}, self.plugin.get_site_profile())
        fake_profile = {
            plugin.KEY_OAUTH_TOKEN: '1',
            plugin.KEY_OAUTH_TOKEN_SECRET: '2',
            plugin.KEY_TWITTER_NAME: '3',
            plugin.KEY_TWITTER_ID: '4',
        }
        self.plugin.save_site_profile('1', '2', '3', '4')
        self.assertEquals(fake_profile, self.plugin.get_site_profile())
        self.plugin.remove_site_profile()
        self.assertEquals({}, self.plugin.get_site_profile())
        self.plugin.save_site_profile('1', '2', '3', '4')

if __name__ == '__main__':
  unittest.main()



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb RESTful API views."""

import datetime
from functools import wraps
import logging
import sys
import traceback
import types

from django.conf import settings
from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.contrib.auth.forms import AuthenticationForm
from django.db.utils import IntegrityError
from django.utils import timezone

from django.http import Http404
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.cache import never_cache
from django.views.decorators.http import require_http_methods
from django.db import transaction
from django.db.models.query import QuerySet

from kegbot.api import kbapi
from kegbot.util import kbjson

from pykeg.contrib.soundserver import models as soundserver_models
from pykeg.core import backend
from pykeg.core import models
from pykeg.proto import protolib
from pykeg.web.api import forms
from pykeg.web.api import util
from pykeg.web.kegadmin.forms import ChangeKegForm

from pykeg.web import tasks

_LOGGER = logging.getLogger(__name__)

### Decorators

def auth_required(viewfunc):
  util.set_needs_auth(viewfunc)
  return viewfunc

### Helpers

def _form_errors(form):
  ret = {}
  for field in form:
    if field.errors:
      name = field.html_name
      ret[name] = []
      for error in field.errors:
        ret[name].append(error)
  return ret

### Endpoints

def all_kegs(request):
  return models.Keg.objects.all().order_by('-start_time')

def all_drinks(request, limit=100):
  qs = models.Drink.objects.all()
  if 'start' in request.GET:
    try:
      start = int(request.GET['start'])
      qs = qs.filter(id__lte=start)
    except ValueError:
      pass
  qs = qs.order_by('-id')
  qs = qs[:limit]
  start = qs[0].id
  return qs

def get_drink(request, drink_id):
  drink = get_object_or_404(models.Drink, id=drink_id)
  return protolib.ToProto(drink, full=True)

@csrf_exempt
@auth_required
def add_drink_photo(request, drink_id):
  if request.method != 'POST':
    raise Http404('Method not supported')
  drink = get_object_or_404(models.Drink, id=drink_id)
  pic = models.Picture.objects.create(image=request.FILES['photo'])
  pour_pic = models.PourPicture.objects.create(picture_id=pic.id,
      drink=drink,
      user=drink.user,
      keg=drink.keg,
      session=drink.session)
  tasks.handle_new_picture.delay(pour_pic.id)
  return protolib.ToProto(pour_pic, full=True)

def get_session(request, session_id):
  session = get_object_or_404(models.DrinkingSession, id=session_id)
  return protolib.ToProto(session, full=True)

def get_session_stats(request, session_id):
  session = get_object_or_404(models.DrinkingSession, id=session_id)
  return session.GetStats()

def get_keg(request, keg_id):
  keg = get_object_or_404(models.Keg, id=keg_id)
  return protolib.ToProto(keg, full=True)

def get_keg_drinks(request, keg_id):
  keg = get_object_or_404(models.Keg, id=keg_id)
  return keg.drinks.all()

def get_keg_events(request, keg_id):
  keg = get_object_or_404(models.Keg, id=keg_id)
  events = keg.events.all()
  events = apply_since(request, events)
  return events

def get_keg_sizes(request):
  return models.KegSize.objects.all()

@require_http_methods(["POST"])
@auth_required
@csrf_exempt
def end_keg(request, keg_id):
  keg = get_object_or_404(models.Keg, id=keg_id)
  tap = keg.current_tap
  keg = request.backend.EndKeg(tap)
  return protolib.ToProto(keg, full=True)

def all_sessions(request):
  return models.DrinkingSession.objects.all()

def current_session(request):
  try:
    latest = models.DrinkingSession.objects.latest()
    if latest.IsActiveNow():
      return latest
  except models.DrinkingSession.DoesNotExist:
    raise Http404

def all_events(request):
  events = models.SystemEvent.objects.all().order_by('-id')
  events = apply_since(request, events)
  events = events[:10]
  return [protolib.ToProto(e, full=True) for e in events]

def apply_since(request, query):
  """Restricts the query to `since` events, if given."""
  since_str = request.GET.get('since')
  if since_str:
    try:
      since = int(since_str)
      return query.filter(id__gt=since)
    except (ValueError, TypeError):
      pass
  return query

@auth_required
def all_sound_events(request):
  return soundserver_models.SoundEvent.objects.all()

def get_keg_sessions(request, keg_id):
  keg = get_object_or_404(models.Keg, id=keg_id)
  sessions = [c.session for c in keg.keg_session_chunks.all()]
  return sessions

def get_keg_stats(request, keg_id):
  keg = get_object_or_404(models.Keg, id=keg_id)
  return keg.GetStats()

def get_system_stats(request):
  return models.KegbotSite.get().GetStats()

def all_taps(request):
  return models.KegTap.objects.all().order_by('name')

@auth_required
def user_list(request):
  return models.User.objects.filter(is_active=True).order_by('username')

def get_user(request, username):
  user = get_object_or_404(models.User, username=username)
  return protolib.ToProto(user, full=True)

def get_user_drinks(request, username):
  user = get_object_or_404(models.User, username=username)
  return user.drinks.all()

def get_user_events(request, username):
  user = get_object_or_404(models.User, username=username)
  return user.events.all()

def get_user_stats(request, username):
  user = get_object_or_404(models.User, username=username)
  return user.get_profile().GetStats()

@auth_required
def get_auth_token(request, auth_device, token_value):
  tok = request.backend.GetAuthToken(auth_device, token_value)
  return tok

@csrf_exempt
@auth_required
def assign_auth_token(request, auth_device, token_value):
  if not request.POST:
    raise kbapi.BadRequestError('POST required.')

  form = forms.AssignTokenForm(request.POST)
  if not form.is_valid():
    errors = _form_errors(form)
    raise kbapi.BadRequestError(errors)

  b = request.backend
  username = form.cleaned_data['username']

  user = backend.get_user(username)
  if not user:
    raise kbapi.BadRequestError("User does not exist")

  try:
    tok = b.GetAuthToken(auth_device, token_value)
  except backend.NoTokenError:
    tok = b.CreateAuthToken(auth_device, token_value, username=username)

  if tok.user != user:
    if tok.user:
      raise kbapi.BadRequestError("Token is already bound to a user")
    tok.user = user
    tok.save()
  return tok

def all_thermo_sensors(request):
  return models.ThermoSensor.objects.all()

def _get_sensor_or_404(request, sensor_name):
  try:
    sensor = models.ThermoSensor.objects.get(raw_name=sensor_name)
  except models.ThermoSensor.DoesNotExist:
    try:
      sensor = models.ThermoSensor.objects.get(nice_name=sensor_name)
    except models.ThermoSensor.DoesNotExist:
      raise Http404
  return sensor

@csrf_exempt
def get_thermo_sensor(request, sensor_name):
  if request.method == 'POST':
    return _thermo_sensor_post(request, sensor_name)
  else:
    return _thermo_sensor_get(request, sensor_name)

def _thermo_sensor_get(request, sensor_name):
  sensor = _get_sensor_or_404(request, sensor_name)
  logs = sensor.thermolog_set.all()
  if not logs:
    last_temp = None
    last_time = None
  else:
    last_temp = logs[0].temp
    last_time = logs[0].time
  res = {
    'sensor': to_dict(sensor),
    'last_temp': last_temp,
    'last_time': last_time,
  }
  return res

@auth_required
def _thermo_sensor_post(request, sensor_name):
  form = forms.ThermoPostForm(request.POST)
  if not form.is_valid():
    raise kbapi.BadRequestError, _form_errors(form)
  cd = form.cleaned_data
  sensor, created = models.ThermoSensor.objects.get_or_create(raw_name=sensor_name)
  # TODO(mikey): use form fields to compute `when`
  return request.backend.LogSensorReading(sensor.raw_name, cd['temp_c'])

def get_thermo_sensor_logs(request, sensor_name):
  sensor = _get_sensor_or_404(request, sensor_name)
  return sensor.thermolog_set.all()[:60*2]

def get_api_key(request):
  user = request.user
  api_key = ''
  if user and (user.is_staff or user.is_superuser):
    api_key = user.get_profile().GetApiKey()
  return {'api_key': api_key}

@csrf_exempt
def tap_detail(request, tap_id):
  tap = get_object_or_404(models.KegTap, meter_name=tap_id)
  if request.method == 'POST':
    return _tap_detail_post(request, tap)
  elif request.method == 'GET':
    return _tap_detail_get(request, tap)
  else:
   raise kbapi.BadRequestError('Method not supported')

def _tap_detail_get(request, tap):
  return protolib.ToProto(tap, full=True)

@csrf_exempt
@auth_required
def tap_calibrate(request, tap_id):
  # TODO(mikey): This would make more semantic sense as PATCH /taps/tap-name/,
  # but Django's support for non-POST verbs is poor (specifically wrt request
  # body/form handling).
  tap = get_object_or_404(models.KegTap, meter_name=tap_id)
  form = forms.CalibrateTapForm(request.POST)
  if form.is_valid():
    tap.ml_per_tick = form.cleaned_data['ml_per_tick']
    tap.save()
  else:
    raise kbapi.BadRequestError, _form_errors(form)
  return protolib.ToProto(tap, full=True)

@csrf_exempt
@auth_required
def tap_spill(request, tap_id):
  tap = get_object_or_404(models.KegTap, meter_name=tap_id)
  if not tap.current_keg:
    raise kbapi.BadRequestError('No keg on tap.')
  form = forms.TapSpillForm(request.POST)
  if form.is_valid():
    tap.current_keg.spilled_ml += form.cleaned_data['volume_ml']
    tap.current_keg.save()
  else:
    raise kbapi.BadRequestError, _form_errors(form)
  return protolib.ToProto(tap, full=True)

@csrf_exempt
@auth_required
def tap_activate(request, tap_id):
  tap = get_object_or_404(models.KegTap, meter_name=tap_id)
  form = ChangeKegForm(request.POST)
  if form.is_valid():
    form.save(tap)
  else:
    raise kbapi.BadRequestError, _form_errors(form)
  return protolib.ToProto(tap, full=True)

@auth_required
def _tap_detail_post(request, tap):
  form = forms.DrinkPostForm(request.POST)
  if not form.is_valid():
    raise kbapi.BadRequestError, _form_errors(form)
  cd = form.cleaned_data
  if cd.get('pour_time') and cd.get('now'):
    pour_time = datetime.datetime.fromtimestamp(cd.get('pour_time'))
    pour_now = datetime.datetime.fromtimestamp(cd.get('now'))
    pour_time_ago = pour_now - pour_time
    pour_time = timezone.now() - pour_time_ago
  else:
    pour_time = None
  duration = cd.get('duration')
  if duration is None:
    duration = 0
  try:
    res = request.backend.RecordDrink(tap_name=tap.meter_name,
      ticks=cd['ticks'],
      volume_ml=cd.get('volume_ml'),
      username=cd.get('username'),
      pour_time=pour_time,
      duration=duration,
      shout=cd.get('shout'),
      tick_time_series=cd.get('tick_time_series'))
    return protolib.ToProto(res, full=True)
  except backend.BackendError, e:
    raise kbapi.ServerError(str(e))

@csrf_exempt
@auth_required
def cancel_drink(request):
  if request.method != 'POST':
    raise kbapi.BadRequestError('POST required')
  form = forms.CancelDrinkForm(request.POST)
  if not form.is_valid():
    raise kbapi.BadRequestError, _form_errors(form)
  cd = form.cleaned_data
  try:
    res = request.backend.CancelDrink(drink_id=cd.get('id'), spilled=cd.get('spilled', False))
    return protolib.ToProto(res, full=True)
  except backend.BackendError, e:
    raise kbapi.ServerError(str(e))

@csrf_exempt
def login(request):
  if request.POST:
    form = AuthenticationForm(data=request.POST)
    if form.is_valid():
      auth_login(request, form.get_user())
      if request.session.test_cookie_worked():
        request.session.delete_test_cookie()
      return {'result': 'ok'}
    else:
      raise kbapi.PermissionDeniedError('Login failed.')
  raise kbapi.BadRequestError('POST required.')

def logout(request):
  auth_logout(request)
  return {'result': 'ok'}

@csrf_exempt
@auth_required
def register(request):
  if not request.POST:
    raise kbapi.BadRequestError('POST required.')
  form = forms.RegisterForm(request.POST)
  errors = {}
  if not form.is_valid():
    errors = _form_errors(form)
  else:
    username = form.cleaned_data['username']
    try:
      u = models.User()
      u.username = username
      u.email = form.cleaned_data['email']
      u.set_password(form.cleaned_data['password'])
      u.save()
      if 'photo' in request.FILES:
        pic = models.Picture.objects.create()
        photo = request.FILES['photo']
        pic.image.save(photo.name, photo)
        pic.save()
        profile = u.get_profile()
        profile.mugshot = pic
        profile.save()
      return protolib.ToProto(u, full=True)
    except IntegrityError:
      user_errs = errors.get('username', [])
      user_errs.append('Username not available.')
      errors['username'] = user_errs
  raise kbapi.BadRequestError(errors)

def default_handler(request):
  raise Http404, "Not an API endpoint: %s" % request.path[:100]



from django.conf import settings

import pykeg
from pykeg.core import features
from pykeg.core import models
from pykeg.web.kegweb.forms import LoginForm

def enabled_features(request):
  """Adds a USE_FEATURENAME flags for each enabled feature (see features.py)"""
  # TODO(mikey): this might make it harder to diagnose why features aren't
  # visible/being used.
  ret = {}
  ret['USE_FACEBOOK'] = features.use_facebook()
  return ret

def kbsite(request):
  kbsite = getattr(request, 'kbsite', None)

  ret = {
    'DEBUG': settings.DEBUG,
    'EMBEDDED': getattr(settings, 'EMBEDDED', False),
    'EPOCH': pykeg.EPOCH,
    'VERSION': pykeg.__version__,
    'HAVE_SESSIONS': False,
    'HAVE_ADMIN': settings.KEGBOT_ENABLE_ADMIN,
    'GOOGLE_ANALYTICS_ID': None,
    'kbsite': kbsite,
    'request_path': request.path,
    'login_form': LoginForm(initial={'next_page': request.path}),
    'guest_info': {
      'name': 'guest',
      'image': None,
    },
    'PLUGINS': getattr(request, 'plugins', {}),
  }

  if kbsite:
    ret['guest_info']['name'] = kbsite.settings.guest_name
    ret['guest_info']['image'] = kbsite.settings.guest_image
    ret['SERIAL_NUMBER'] = kbsite.serial_number
    ret['HAVE_SESSIONS'] = models.DrinkingSession.objects.all().count() > 0
    ret['GOOGLE_ANALYTICS_ID'] = kbsite.settings.google_analytics_id

  return ret
